# Usage: ls-ingest <dataset_id>
import os
import json
import argparse

from latentscope.util import get_data_dir
from latentscope import __version__

# TODO make a parquet version of these
def main():
    parser = argparse.ArgumentParser(description='Ingest a dataset')
    parser.add_argument('id', type=str, help='Dataset id (directory name in data folder)')
    parser.add_argument('--path', type=str, help='Path to csv/parquet/json/jsonl/xlsx file, otherwise assumes input.csv in dataset directory')
    parser.add_argument('--text_column', type=str, help='Column to use as text for the scope')
    args = parser.parse_args()
    ingest_file(args.id, args.path, args.text_column)

def ingest_file(dataset_id, file_path, text_column = None):
    import pandas as pd
    DATA_DIR = get_data_dir()
    directory = os.path.join(DATA_DIR, dataset_id)
    # check if dataset exists, if it does we want to increment a postfix on the dataset_id
    # if os.path.exists(directory):
    #     postfix = 1
    #     while os.path.exists(f"{directory}-{postfix:03d}"):
    #         postfix += 1
    #     dataset_id = f"{dataset_id}-{postfix:03d}"
    #     directory = os.path.join(DATA_DIR, dataset_id)
    # os.makedirs(directory)
    print("RUNNING:", dataset_id)

    if not file_path:
        file_path = os.path.join(directory, "input.csv")
    file_type = file_path.split('.')[-1]
    print(f"File type detected: {file_type}")
    file = os.path.join(file_path)
    print("reading", file)
    if file_type == "csv":
        df = pd.read_csv(file)
    elif file_type == "parquet":
        df = pd.read_parquet(file)
    elif file_type == "jsonl":
        df = pd.read_json(file, lines=True)
    elif file_type == "json":
        df = pd.read_json(file)
    elif file_type == "xlsx":
        df = pd.read_excel(file)
    else:
        raise ValueError(f"Unsupported file type: {file_type}")

    ingest(dataset_id, df, text_column)


def ingest(dataset_id, df, text_column = None):


    DATA_DIR = get_data_dir()
    print("DATA DIR", DATA_DIR)
    directory = os.path.join(DATA_DIR, dataset_id)
    print("DIRECTORY", directory)
    if not os.path.exists(directory):
        os.makedirs(directory)
    df = df.reset_index(drop=True)

    print(df.head())
    print(df.tail())
    print(df.columns)

    print("checking column types")
    import pandas as pd
    import numpy as np
    # determine the types of the values in columns, especially string, number or array of numbers
    # we will store these in the metadata
    # we will also store the number of unique values in each column
    column_metadata = {}
    for column in df.columns:
        non_null_series = df[column].dropna()
        # TODO: handle dates even if they are strings
        if pd.api.types.is_datetime64_any_dtype(non_null_series):
            column_type = "date"
        elif pd.api.types.is_string_dtype(non_null_series):
            column_type = "string"
        elif pd.api.types.is_numeric_dtype(non_null_series):
            column_type = "number"
        elif all(non_null_series.apply(lambda x: isinstance(x, list) and all(isinstance(i, (int, float)) for i in x))):
            print("list array of numbers", column)
            column_type = "array"
        elif isinstance(non_null_series.iloc[0], np.ndarray):
            print("np array", column)
            column_type = "array"
        else:
            column_type = "unknown"

        if column_type == "unknown":
            print("unknown column type", column, "converting to string")
            df[column] = df[column].astype(str)
            column_type = "string"
        
        print("COLUMN", column, "TYPE", column_type)

        # Count unique values, excluding NaN
        try:
            if isinstance(df[column].iloc[0], np.ndarray):
                unique_values_count = len(set([tuple(x) for x in df[column].dropna()]))
            elif isinstance(df[column].iloc[0], bytes):
                unique_values_count = len(set(df[column].dropna().apply(lambda x: x.decode('utf-8'))))
            elif isinstance(df[column].iloc[0], dict):
                unique_values_count = len(set(df[column].dropna().apply(json.dumps)))
            else:
                unique_values_count = df[column].nunique(dropna=True)
        except:
            unique_values_count = -1

        # Store the metadata
        column_metadata[column] = {
            "type": column_type,
            "unique_values_count": unique_values_count
        }
        if column_type == "string" and unique_values_count <= 100:
            categories = df[column].value_counts().index.tolist()
            column_metadata[column]["categories"] = categories
            column_metadata[column]["counts"] = df[column].value_counts().to_dict()
        if column_type == "string":
            if df[column].str.startswith("http").all():
                column_metadata[column]["url"] = True
                # check if endings of string are common image formats like png, jpg, jpeg, webp
                if df[column].str.lower().str.endswith(("png", "jpg", "jpeg", "webp", "svg", "gif")).all():
                    column_metadata[column]["image"] = True
        if column_type == "number":
            extent = df[column].agg(['min', 'max'])
            column_metadata[column]["extent"] = [None if pd.isna(x) else x for x in extent.tolist()]
        if column_type == "date":
            extent = df[column].agg(['min', 'max'])
            column_metadata[column]["extent"] = extent.tolist()

    output_file = f"{directory}/input.parquet"
    df.to_parquet(output_file)
    print("wrote", output_file)

    # write out a json file with the model name and shape of the embeddings
    if text_column is None:
        text_column = "text" if "text" in df.columns else None
    if text_column is None:
        text_column = next((col for col, meta in column_metadata.items() if meta['type'] == 'string'), None)

    potential_embeddings = [col for col, meta in column_metadata.items() if meta['type'] == 'array']
    with open(os.path.join(directory,'meta.json'), 'w') as f:
        json.dump({
            "id": dataset_id,
            "length": df.shape[0],
            "columns": df.columns.tolist(),
            "text_column": text_column,
            "column_metadata": column_metadata,
            "potential_embeddings": potential_embeddings,
            "ls_version": __version__
            }, f, indent=2)

    # create all the directories we will use
    os.makedirs(os.path.join(DATA_DIR, dataset_id, "embeddings"), exist_ok=True)
    os.makedirs(os.path.join(DATA_DIR, dataset_id, "saes"), exist_ok=True)
    os.makedirs(os.path.join(DATA_DIR, dataset_id, "umaps"), exist_ok=True)
    os.makedirs(os.path.join(DATA_DIR, dataset_id, "clusters"), exist_ok=True)
    os.makedirs(os.path.join(DATA_DIR, dataset_id, "scopes"), exist_ok=True)
    tags_dir = os.path.join(DATA_DIR, dataset_id, "tags")
    os.makedirs(tags_dir, exist_ok=True)
    # Create default tags if they don't exist
    new_files = ['👍.indices', '👎.indices']
    for file_name in new_files:
        file_path = os.path.join(tags_dir, file_name)
        if not os.path.exists(file_path):
            with open(file_path, 'w') as file:
                file.write('')  # Initialize with an empty JSON object


if __name__ == "__main__":
    main()



# Usage: ls-label <dataset_id> <text_column> <cluster_id> <model_id> <context>
import os
import re
import sys
import json
import time
import argparse
from datetime import datetime

try:
    # Check if the runtime environment is a Jupyter notebook
    if 'ipykernel' in sys.modules and 'IPython' in sys.modules:
        from tqdm.notebook import tqdm
    else:
        from tqdm import tqdm
except ImportError as e:
    # Fallback to the standard console version if import fails
    from tqdm import tqdm

from latentscope.util import get_data_dir
from latentscope.models import get_chat_model

def chunked_iterable(iterable, size):
    """Yield successive chunks from an iterable."""
    for i in range(0, len(iterable), size):
        yield iterable[i:i + size]

def too_many_duplicates(line, threshold=100):
    word_count = {}
    if not line:
        return False
    words = str(line).split()
    for word in words:
        word_count[word] = word_count.get(word, 0) + 1
    return any(count > threshold for count in word_count.values())

def main():
    parser = argparse.ArgumentParser(description='Label a set of slides using OpenAI')
    parser.add_argument('dataset_id', type=str, help='Dataset ID (directory name in data/)')
    parser.add_argument('text_column', type=str, help='Output file', default='text')
    parser.add_argument('cluster_id', type=str, help='ID of cluster set', default='cluster-001')
    parser.add_argument('model_id', type=str, help='ID of model to use', default="openai-gpt-3.5-turbo")
    parser.add_argument('samples', type=int, help='Number to sample from each cluster (default: 0 for all)', default=0)
    parser.add_argument('context', type=str, help='Additional context for labeling model', default="")
    parser.add_argument('--rerun', type=str, help='Rerun the given embedding from last completed batch')
    parser.add_argument('--max_tokens_per_sample', type=int, help='Max tokens per sample', default=-1)
    parser.add_argument('--max_tokens_total', type=int, help='Max tokens total', default=-1)

    # Parse arguments
    args = parser.parse_args()

    labeler(args.dataset_id, args.text_column, args.cluster_id, args.model_id, args.samples, args.context, args.rerun, args.max_tokens_per_sample, args.max_tokens_total)


def labeler(dataset_id, text_column="text", cluster_id="cluster-001", model_id="openai-gpt-3.5-turbo", samples=0, context="", rerun="", max_tokens_per_sample=-1, max_tokens_total=-1):
    import numpy as np
    import pandas as pd
    DATA_DIR = get_data_dir()
    df = pd.read_parquet(os.path.join(DATA_DIR, dataset_id, "input.parquet"))

    # Load the indices for each cluster from the prepopulated labels file generated by cluster.py
    cluster_dir = os.path.join(DATA_DIR, dataset_id, "clusters")
    clusters = pd.read_parquet(os.path.join(cluster_dir, f"{cluster_id}-labels-default.parquet"))
    # initialize the labeled property to false when loading default clusters
    clusters = clusters.copy()
    clusters['labeled'] = False
    
    cluster_rows = pd.read_parquet(os.path.join(cluster_dir, f"{cluster_id}.parquet"))
    df["cluster"] = cluster_rows["cluster"]
    df["raw_cluster"] = cluster_rows["raw_cluster"]

    with open(os.path.join(cluster_dir, f"{cluster_id}.json"), 'r') as f:
        cluster_meta = json.load(f)
    umap_id = cluster_meta["umap_id"]
    umap = pd.read_parquet(os.path.join(DATA_DIR, dataset_id, "umaps", f"{umap_id}.parquet"))
    df["x"] = umap["x"]
    df["y"] = umap["y"]

    unlabeled_row = 0
    if rerun is not None:
        label_id = rerun
        # print(clusters.columns)
        # find the first row where labeled isnt True
        unlabeled_row = clusters[~clusters['labeled']].first_valid_index()
        tqdm.write(f"First unlabeled row: {unlabeled_row}")
        

    else:
        # Determine the label id for the given cluster_id by checking existing label files
        label_files = [f for f in os.listdir(cluster_dir) if re.match(rf"{re.escape(cluster_id)}-labels-\d+\.parquet", f)]
        if label_files:
            # Extract label numbers and find the maximum
            label_numbers = [int(re.search(rf"{re.escape(cluster_id)}-labels-(\d+)\.parquet", f).group(1)) for f in label_files]
            next_label_number = max(label_numbers) + 1
        else:
            next_label_number = 1
        label_id = f"{cluster_id}-labels-{next_label_number:03d}"

    # track history of model_id used
    history_file_path = os.path.join(DATA_DIR, "chat_model_history.csv")
    try:
        with open(history_file_path, 'a') as history_file:
            history_file.write(f"{datetime.now().isoformat()},{model_id}\n")
    except FileNotFoundError:
        with open(history_file_path, 'w') as history_file:
            history_file.write(f"{datetime.now().isoformat()},{model_id}\n")

    tqdm.write(f"RUNNING: {label_id}")

    tqdm.write(f"Loading model {model_id} (may take a while if first time downloading from HF)")
    model = get_chat_model(model_id)
    model.load_model()
    enc = model.encoder
    tqdm.write(f"Model loaded")

    # unescape the context
    context = context.replace('\\"', '"')

    # Create the lists of items we will send for summarization
    # we truncate the list based on tokens and we also remove items that have too many duplicate words
    extracts = []
    for i, row in tqdm(clusters.iterrows(), total=clusters.shape[0], desc="Preparing extracts"):
    # for i, row in clusters.iterrows():
        indices = row['indices']
        # items = df.loc[list(indices), text_column]
        items = df.loc[list(indices)]
        if samples > 0 and samples < len(items):
            # first sample the items from cluster_rows where 'raw_cluster' matches the current cluster_id
            cluster_items = items[items['raw_cluster'] == i]

            if(len(cluster_items) < samples):
                cluster_items = pd.concat([cluster_items, items[items['cluster'] == i]])

            # Sort cluster items by distance from centroid
            # Get x,y coordinates for items
            coords = cluster_items[['x', 'y']].values
            
            # Calculate centroid
            centroid = coords.mean(axis=0)
            
            # Calculate distances from centroid
            distances = np.sqrt(np.sum((coords - centroid) ** 2, axis=1))
            
            # Add distances as column and sort
            cluster_items = cluster_items.assign(centroid_dist=distances)
            cluster_items = cluster_items.sort_values('centroid_dist')

            items = cluster_items[0:samples]
            # items = cluster_items.sample(samples)

        items = items.drop_duplicates()
        items = items[text_column]
        # tqdm.write(f"{i} items: {len(items)}")
        
        total_tokens = 0
        keep_items = []
        if enc is not None:
            for item in items:
                if item is None:
                    continue
                encoded_item = enc.encode(item)
                if max_tokens_per_sample > 0 and len(encoded_item) > max_tokens_per_sample:
                    item = enc.decode(encoded_item[:max_tokens_per_sample])
                    total_tokens += max_tokens_per_sample
                else:
                    total_tokens += len(encoded_item)
                if max_tokens_total > 0 and total_tokens > max_tokens_total:
                    break
                # tqdm.write(f"tokens: {len(encoded_item)}")
                keep_items.append(item)
        else:
            keep_items = items
        keep_items = [item for item in keep_items if item is not None]
        # tqdm.write(f"{i} total tokens: {total_tokens}, keep_items: {len(keep_items)}")
        extracts.append(keep_items)

        # text = '\n'.join([f"{i+1}. {t}" for i, t in enumerate(items) if not too_many_duplicates(t)])
        # text = '\n'.join([f"<ListItem>{t}</ListItem>" for i, t in enumerate(items) if not too_many_duplicates(t)])
        # encoded_text = enc.encode(text)
        # if len(encoded_text) > max_tokens:
        #     encoded_text = encoded_text[:max_tokens]
        # extract = enc.decode(encoded_text)
        # extracts.append(extract)

    # TODO we arent really batching these
    # each "batch" is the items for a single cluster
    batch_size = 1
    labels = []
    clean_labels = []

    for i,batch in enumerate(tqdm(chunked_iterable(extracts, batch_size),  total=len(extracts)//batch_size)):
        # tqdm.write(batch[0])
        if(unlabeled_row > 0):
            if clusters.loc[i, 'labeled']:
                tqdm.write(f"skipping {i} already labeled {clusters.loc[i, 'label']}")
                time.sleep(0.01)
                continue

        try:
            time.sleep(0.01)
            # tqdm.write(f"Summarizing {batch[0]}")
            label = model.summarize(batch[0], context)
            labels.append(label)
            
            # do some cleanup of the labels when the model doesn't follow instructions
            clean_label = label.replace("\n", " ")
            clean_label = clean_label.replace("<|eot_id|>", "")
            clean_label = clean_label.replace('*', '')
            clean_label = clean_label.replace('"', '')
            clean_label = clean_label.replace("'", '')
            # clean_label = clean_label.replace("-", '')
            clean_label = ' '.join(clean_label.split())
            clean_label = " ".join(clean_label.split(" ")[0:5])
            clean_labels.append(clean_label)
            if re.search(r"please provide", label, re.IGNORECASE):
                tqdm.write(f"cluster {i} label: {clean_label}")
                tqdm.write(f"batch: {batch[0]}")
                tqdm.write(f"label: {label}")
            
            tqdm.write(f"cluster {i} label: {clean_label}")
            clusters.loc[i, 'label'] = clean_label
            clusters.loc[i, 'label_raw'] = label
            clusters.loc[i, 'labeled'] = True
            # length = len(clean_labels) - 1
            # clusters_df.loc[unlabled_row:unlabled_row+length, 'label'] = clean_labels
            # clusters_df.loc[unlabled_row:unlabled_row+length, 'label_raw'] = labels
            # clusters_df.loc[unlabled_row:unlabled_row+length, 'labeled'] = [True for i in range(0, len(labels))]
            clusters.to_parquet(os.path.join(cluster_dir, f"{label_id}.parquet"))
            # update 

        except Exception as e: 
            tqdm.write(f"{batch[0]}")
            tqdm.write(f"ERROR: {e}")
            tqdm.write("exiting")
            exit(1)

    print("labels:", len(labels))
    # add lables to slides df

    # write the df to parquet
    with open(os.path.join(cluster_dir,f"{label_id}.json"), 'w') as f:
        json.dump({
            "id": label_id,
            "cluster_id": cluster_id,
            "model_id": model_id, 
            "text_column": text_column,
            "samples": samples,
            "context": context,
            # "system_prompt": system_prompt,
            "max_tokens_per_sample": max_tokens_per_sample,
            "max_tokens_total": max_tokens_total,
        }, f, indent=2)
    f.close()
    print("done with", label_id)

if __name__ == "__main__":
    main()



# Usage: python umapper.py <dataset_id> <model> <neighbors> <min_dist>
# Example: python umapper.py dadabase-curated BAAI_bge-small-en-v1.5 50 0.075
import os
import re
import sys
import json
import argparse

from latentscope.util import get_data_dir

def main():
    parser = argparse.ArgumentParser(description='UMAP embeddings for a dataset')
    parser.add_argument('dataset_id', type=str, help='Dataset name (directory name in data/)')
    parser.add_argument('embedding_id', type=str, help='Name of embedding model to use')
    parser.add_argument('neighbors', type=int, nargs="?", help='Output file', default=25)
    parser.add_argument('min_dist', type=float, nargs="?", help='Output file', default=0.075)
    parser.add_argument('--init', type=str, help='Initialize with UMAP', default=None)
    parser.add_argument('--align', type=str, help='Align UMAP with multiple embeddings', default=None)
    parser.add_argument('--save', action='store_true', help='Save the UMAP model')
    parser.add_argument('--seed', type=int, help='Random seed', default=None)
    parser.add_argument('--sae_id', type=str, help='SAE to project instead of embedding', default=None)

    # Parse arguments
    args = parser.parse_args()

    seed = args.seed
    if seed == -1:
        seed = None

    if args.sae_id:
        sparse_umapper(args.dataset_id, args.embedding_id, args.sae_id, args.neighbors, args.min_dist, save=args.save, init=args.init, seed=seed)
    else:
        umapper(args.dataset_id, args.embedding_id, args.neighbors, args.min_dist, save=args.save, init=args.init, align=args.align, seed=seed)


# TODO move this into shared space
def calculate_point_size(num_points, min_size=10, max_size=30, base_num_points=100):
    import numpy as np
    """
    Calculate the size of points for a scatter plot based on the number of points.
    """
    # TODO fix this to actually calculate a log scale between min and max size
    if num_points <= base_num_points:
        return max_size
    else:
        return min(min_size + min_size * np.log(num_points / base_num_points), max_size)

def load_embeddings(dataset_id, embedding_id):
    import h5py
    import numpy as np
    DATA_DIR = get_data_dir()

    if embedding_id[0:3] == "sae":
        sae_path = os.path.join(DATA_DIR, dataset_id, "saes", f"{embedding_id}.h5")
        with h5py.File(sae_path, 'r') as f:
            all_acts = f.get("top_acts")[:]
            all_indices = f.get("top_indices")[:]

        # read the meta json for sae
        with open(os.path.join(DATA_DIR, dataset_id, "saes", f"{embedding_id}.json"), 'r') as f:
            meta = json.load(f)
        # print("SAE META", meta["id"], meta["num_features"], "features", meta["model_id"], meta["k_expansion"])
        
        import scipy
        # print("ALL ACTS SHAPE", all_acts.shape)
        # print("ALL INDS SHAPE", all_indices.shape)
        matrix = scipy.sparse.lil_matrix((all_acts.shape[0], meta["num_features"]), dtype=np.float32)
        for i in range(all_acts.shape[0]):
            matrix.rows[i] = all_indices[i].tolist()
            matrix.data[i] = all_acts[i].tolist()
        return matrix
    else:
        emb_path = os.path.join(DATA_DIR, dataset_id, "embeddings", f"{embedding_id}.h5")
        with h5py.File(emb_path, 'r') as f:
            dataset = f["embeddings"]
            embeddings = np.array(dataset)
            return embeddings

def umapper(dataset_id, embedding_id, neighbors=25, min_dist=0.1, save=False, init=None, align=None, seed=None):
    DATA_DIR = get_data_dir()
    # read in the embeddings 

    umap_dir = os.path.join(DATA_DIR, dataset_id, "umaps")
    if not os.path.exists(umap_dir):
        os.makedirs(umap_dir)

    # determine the index of the last umap run by looking in the dataset directory
    # for files named umap-<number>.json
    umap_files = [f for f in os.listdir(umap_dir) if re.match(r"umap-\d+\.json", f)]
    if len(umap_files) > 0:
        last_umap = sorted(umap_files)[-1]
        last_umap_number = int(last_umap.split("-")[1].split(".")[0])
        next_umap_number = last_umap_number + 1
    else:
        next_umap_number = 1

    # make the umap name from the number, zero padded to 3 digits
    umap_id = f"umap-{next_umap_number:03d}"
    print("RUNNING:", umap_id)

    import umap
    import h5py
    import pickle
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt

    print("loading embeddings")
    # embedding_path = os.path.join(DATA_DIR, dataset_id, "embeddings", f"{embedding_id}.h5")
    # with h5py.File(embedding_path, 'r') as f:
    #     dataset = f["embeddings"]
    #     embeddings = np.array(dataset)
    embeddings = load_embeddings(dataset_id, embedding_id)

    def process_umap_embeddings(umap_id, umap_embeddings, emb_id, align_id=None):
        min_values = np.min(umap_embeddings, axis=0)
        max_values = np.max(umap_embeddings, axis=0)

        # Scale the embeddings to the range [0, 1]
        umap_embeddings = (umap_embeddings - min_values) / (max_values - min_values)

        # Scale the embeddings to the range [-1, 1]
        umap_embeddings = 2 * umap_embeddings - 1

        print("writing normalized umap", umap_id)
        # save umap embeddings to a parquet file with columns x,y
        df = pd.DataFrame(umap_embeddings, columns=['x', 'y'])

        # TODO I moved this to scope.py it's cheap and it makes it backwards compatible
        # Calculate tile indices for a 64x64 grid from -1 to 1
        # def make_tiles(x, y, num_tiles=64):
        #     tile_size = 2.0 / num_tiles  # Size of each tile (-1 to 1 = range of 2)
            
        #     # Calculate row and column indices (0-63) for each point
        #     col_indices = np.floor((x + 1) / tile_size).astype(int)
        #     row_indices = np.floor((y + 1) / tile_size).astype(int)
            
        #     # Clip indices to valid range in case of numerical edge cases
        #     col_indices = np.clip(col_indices, 0, num_tiles - 1)
        #     row_indices = np.clip(row_indices, 0, num_tiles - 1)
            
        #     # Convert 2D grid indices to 1D tile index (row * num_cols + col)
        #     tile_indices = row_indices * num_tiles + col_indices
        #     return tile_indices

        # df['tile_index_32'] = make_tiles(umap_embeddings[:, 0], umap_embeddings[:, 1], 32)
        # df['tile_index_64'] = make_tiles(umap_embeddings[:, 0], umap_embeddings[:, 1], 64)
        # df['tile_index_128'] = make_tiles(umap_embeddings[:, 0], umap_embeddings[:, 1], 128)

        output_file = os.path.join(umap_dir, f"{umap_id}.parquet")
        df.to_parquet(output_file)
        print("wrote", output_file)

        # generate a scatterplot of the umap embeddings and save it to a file
        fig, ax = plt.subplots(figsize=(14.22, 14.22))  # 1024px by 1024px at 72 dpi
        point_size = calculate_point_size(umap_embeddings.shape[0])
        print("POINT SIZE", point_size, "for", umap_embeddings.shape[0], "points")
        plt.scatter(umap_embeddings[:, 0], umap_embeddings[:, 1], s=point_size, alpha=0.5)
        plt.axis('off')  # remove axis
        plt.gca().set_position([0, 0, 1, 1])  # remove margins
        plt.savefig(os.path.join(umap_dir, f"{umap_id}.png"))

        # save a json file with the umap parameters
        with open(os.path.join(umap_dir, f'{umap_id}.json'), 'w') as f:
            meta = {
                "id": umap_id, 
                "embedding_id": emb_id,
                "neighbors": neighbors, 
                "min_dist": min_dist,
                "min_values": min_values.tolist(),
                "max_values": max_values.tolist(),
            }
            if init is not None and init != "":
                meta["init"] = init,
            if align is not None and align != "":
                meta["align"] = f"{embedding_id},{align}"
                meta["align_id"] = align_id
            json.dump(meta, f, indent=2)
        f.close()

    ### END OF process_umap_embeddings


    if align is not None and align != "":
        print("aligned umap", align)
        # split the align string into umap names
        embs = align.split(",")
        # load each embedding from its h5 file
        a_embedding_ids = [embedding_id]
        a_embeddings = [embeddings]
        for emb in embs:
            print("loading", emb)
            # emb_path = os.path.join(DATA_DIR, dataset_id, "embeddings", f"{emb}.h5")
            # with h5py.File(emb_path, 'r') as f:
            #     dataset = f["embeddings"]
            #     a_emb = np.array(dataset)
            a_emb = load_embeddings(dataset_id, emb)
            print("loaded", emb, "shape", a_emb.shape)
            a_embeddings.append(a_emb)
            a_embedding_ids.append(emb)
        
        reducer = umap.AlignedUMAP(
            n_neighbors=neighbors,
            min_dist=min_dist,
            metric='cosine',
            random_state=seed,
            n_components=2,
            verbose=True,
        )
        print("a_embeddings", len(a_embeddings), a_embeddings[0].shape[0])
        relations = [{j: j for j in range(a_embeddings[i].shape[0])} for i in range(len(a_embeddings)-1)]
        print("relations", len(relations))
        aligned = reducer.fit_transform(a_embeddings, relations=relations)
        print("ALIGNED", aligned)
        for i,emb in enumerate(a_embedding_ids):
            print("processing", emb, "umap", next_umap_number+i)
            process_umap_embeddings(f"umap-{next_umap_number+i:03d}", aligned[i], emb, umap_id)

        print("done with aligned umap")
        return 

    if init is not None and init != "":
        print("loading umap", init)
        initial_df = pd.read_parquet(os.path.join(umap_dir, f"{init}.parquet"))
        initial = initial_df.to_numpy()
        print("initial shape", initial.shape)
        reducer = umap.UMAP(
            init=initial,
            n_neighbors=neighbors,
            min_dist=min_dist,
            metric='cosine',
            random_state=seed,
            n_components=2,
            verbose=True,
        )
    else:
        reducer = umap.UMAP(
            n_neighbors=neighbors,
            min_dist=min_dist,
            metric='cosine',
            random_state=seed,
            n_components=2,
            verbose=True,
        )
    print("reducing", embeddings.shape[1], "embeddings to 2 dimensions")

    import sys
    sys.stdout.flush()

    umap_embeddings = reducer.fit_transform(embeddings)
    process_umap_embeddings(umap_id, umap_embeddings, embedding_id)

    if save:
        # save a pickle of the umap
        with open(os.path.join(umap_dir, f'{umap_id}.pkl'), 'wb') as f:
            pickle.dump(reducer, f)

    print("done with", umap_id)

def sparse_umapper(dataset_id, embedding_id, sae_id, neighbors=25, min_dist=0.1, save=False, init=None, seed=None):
    DATA_DIR = get_data_dir()
    # read in the embeddings 

    umap_dir = os.path.join(DATA_DIR, dataset_id, "umaps")
    if not os.path.exists(umap_dir):
        os.makedirs(umap_dir)

    # determine the index of the last umap run by looking in the dataset directory
    # for files named umap-<number>.json
    umap_files = [f for f in os.listdir(umap_dir) if re.match(r"umap-\d+\.json", f)]
    if len(umap_files) > 0:
        last_umap = sorted(umap_files)[-1]
        last_umap_number = int(last_umap.split("-")[1].split(".")[0])
        next_umap_number = last_umap_number + 1
    else:
        next_umap_number = 1

    # make the umap name from the number, zero padded to 3 digits
    umap_id = f"umap-{next_umap_number:03d}"
    print("RUNNING:", umap_id)

    import umap
    import h5py
    import pickle
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt

    print("loading sparse embeddings")
    matrix = load_embeddings(dataset_id, sae_id)
    # sae_path = os.path.join(DATA_DIR, dataset_id, "saes", f"{sae_id}.h5")
    # with h5py.File(sae_path, 'r') as f:
    #     all_acts = f.get("top_acts")[:]
    #     all_indices = f.get("top_indices")[:]

    # # read the meta json for sae
    # with open(os.path.join(DATA_DIR, dataset_id, "saes", f"{sae_id}.json"), 'r') as f:
    #     meta = json.load(f)
    # print("SAE META", meta["id"], meta["num_features"], "features", meta["model_id"], meta["k_expansion"])
    
    # import scipy
    # print("ALL ACTS SHAPE", all_acts.shape)
    # print("ALL INDS SHAPE", all_indices.shape)
    # matrix = scipy.sparse.lil_matrix((all_acts.shape[0], meta["num_features"]), dtype=np.float32)
    # # matrix.rows = all_indices.tolist()
    # # matrix.data = all_acts.tolist()
    # for i in range(all_acts.shape[0]):
    #     matrix.rows[i] = all_indices[i].tolist()
    #     matrix.data[i] = all_acts[i].tolist()
    
    def process_umap_embeddings(umap_id, umap_embeddings, emb_id, sae_id, align_id=None):
        min_values = np.min(umap_embeddings, axis=0)
        max_values = np.max(umap_embeddings, axis=0)

        # Scale the embeddings to the range [0, 1]
        umap_embeddings = (umap_embeddings - min_values) / (max_values - min_values)

        # Scale the embeddings to the range [-1, 1]
        umap_embeddings = 2 * umap_embeddings - 1

        print("writing normalized umap", umap_id)
        # save umap embeddings to a parquet file with columns x,y
        df = pd.DataFrame(umap_embeddings, columns=['x', 'y'])
        output_file = os.path.join(umap_dir, f"{umap_id}.parquet")
        df.to_parquet(output_file)
        print("wrote", output_file)

        # generate a scatterplot of the umap embeddings and save it to a file
        fig, ax = plt.subplots(figsize=(14.22, 14.22))  # 1024px by 1024px at 72 dpi
        point_size = calculate_point_size(umap_embeddings.shape[0])
        print("POINT SIZE", point_size, "for", umap_embeddings.shape[0], "points")
        plt.scatter(umap_embeddings[:, 0], umap_embeddings[:, 1], s=point_size, alpha=0.5)
        plt.axis('off')  # remove axis
        plt.gca().set_position([0, 0, 1, 1])  # remove margins
        plt.savefig(os.path.join(umap_dir, f"{umap_id}.png"))

        # save a json file with the umap parameters
        with open(os.path.join(umap_dir, f'{umap_id}.json'), 'w') as f:
            meta = {
                "id": umap_id, 
                "embedding_id": emb_id,
                "sae_id": sae_id,
                "neighbors": neighbors, 
                "min_dist": min_dist,
            }
            if init is not None and init != "":
                meta["init"] = init,
            json.dump(meta, f, indent=2)
        f.close()


    if init is not None and init != "":
        print("loading umap", init)
        initial_df = pd.read_parquet(os.path.join(umap_dir, f"{init}.parquet"))
        initial = initial_df.to_numpy()
        print("initial shape", initial.shape)
        reducer = umap.UMAP(
            init=initial,
            n_neighbors=neighbors,
            min_dist=min_dist,
            metric='cosine',
            random_state=seed,
            n_components=2,
            verbose=True,
        )
    else:
        reducer = umap.UMAP(
            n_neighbors=neighbors,
            min_dist=min_dist,
            metric='cosine',
            random_state=seed,
            n_components=2,
            verbose=True,
        )
    print("reducing", matrix.shape[0], "sparse features to 2 dimensions")
    import sys
    sys.stdout.flush()

    import time
    time.sleep(1)  # Wait for a second before starting
    umap_embeddings = reducer.fit_transform(matrix)
    process_umap_embeddings(umap_id, umap_embeddings, embedding_id, sae_id)

    if save:
        # save a pickle of the umap
        with open(os.path.join(umap_dir, f'{umap_id}.pkl'), 'wb') as f:
            pickle.dump(reducer, f)
    print("done with", umap_id)



if __name__ == "__main__":
    main()



import os
import time
import json
import uuid
import subprocess
import threading
from datetime import datetime
from flask import Blueprint, jsonify, request

# Create a Blueprint
jobs_bp = Blueprint('jobs_bp', __name__)
jobs_write_bp = Blueprint('jobs_write_bp', __name__)
DATA_DIR = os.getenv('LATENT_SCOPE_DATA')

TIMEOUT = 60 * 5 # 5 minute timeout TODO: make this a config option

PROCESSES = {}

def run_job(dataset, job_id, command):
    job_dir = os.path.join(DATA_DIR, dataset, "jobs")
    if not os.path.exists(job_dir):
      os.makedirs(job_dir)

    progress_file = os.path.join(job_dir, f"{job_id}.json")
    print("command", command)
    job_name = command.split(" ")[0]
    if "ls-" in job_name:
        job_name = job_name.replace("ls-", "")
    job = {
        "id": job_id,
        "dataset": dataset,
        "job_name": job_name,
        "command": command, 
        "status": "running", 
        "last_update": str(datetime.now()), 
        "progress": [], 
        "times": []
    }

    with open(progress_file, 'w') as f:
        json.dump(job, f)

    # bufsize 1 flushes the stdout faster 
    env = os.environ.copy()
    env['PYTHONUNBUFFERED'] = '1'
    # TODO: need to watch for exploits in command if using shell=True for security reasons
    process = subprocess.Popen(
        command, 
        stdout=subprocess.PIPE, 
        stderr=subprocess.STDOUT, 
        text=True, 
        shell=True, 
        encoding="utf-8",
        env=env,
        bufsize=1
    )
    PROCESSES[job_id] = process

    last_output_time = time.time()

    # Create thread to handle stderr separately
    # def handle_stderr():
    #     for line in iter(process.stderr.readline, ''):
    #         print("stderr:", line.strip())
    #         job["progress"].append(line.strip())
    #         job["times"].append(str(datetime.now()))
    #         job["last_update"] = str(datetime.now())
    #         with open(progress_file, 'w') as f:
    #             json.dump(job, f)
    #         nonlocal last_output_time
    #         last_output_time = time.time()

    # stderr_thread = threading.Thread(target=handle_stderr)
    # stderr_thread.daemon = True
    # stderr_thread.start()

    while True:
        output = process.stdout.readline()
        current_time = time.time()
        print(current_time, current_time - last_output_time, TIMEOUT)
        print("output", output)

        if output == '' and process.poll() is not None:
            break
        if output:
            print(output.strip())
            if("RUNNING:" in output):
                run_id = output.strip().split("RUNNING: ")[1]
                print("found the id", run_id)
                job["run_id"] = run_id
            job["progress"].append(output.strip())
            job["times"].append(str(datetime.now()))
            job["last_update"] = str(datetime.now())
            with open(progress_file, 'w') as f:
                json.dump(job, f)
            last_output_time = current_time
        elif current_time - last_output_time > TIMEOUT:
            print(f"Timeout: No output for more than {TIMEOUT} seconds.")
            print("OUTPUT", output)
            job["progress"].append(output.strip())
            job["progress"].append(f"Timeout: No output for more than {TIMEOUT} seconds.")
            job["status"] = "error"
            break  # Break the loop

    # stderr_thread.join(timeout=1)  # Wait for stderr thread to finish

    if process.returncode != 0:
        job["status"] = "error"
    else:
        job["status"] = "completed"
    # job["status"] = "completed"

    with open(progress_file, 'w') as f:
        json.dump(job, f)

@jobs_bp.route('/job')
def get_job():
    print("get_job", request.args)
    dataset = request.args.get('dataset')
    job_id = request.args.get('job_id')
    progress_file = os.path.join(DATA_DIR, dataset, "jobs", f"{job_id}.json")
    if os.path.exists(progress_file):
        try:
            with open(progress_file, 'r') as f:
                job = json.load(f)
        except:
            time.sleep(0.1)
            with open(progress_file, 'r') as f:
                job = json.load(f)
        return jsonify(job)
    else:
        return jsonify({'status': 'not found'}), 404

@jobs_bp.route('/all')
def get_jobs():
    dataset = request.args.get('dataset')
    job_dir = os.path.join(DATA_DIR, dataset, "jobs")
    jobs = []
    if os.path.exists(job_dir):
        for file in os.listdir(job_dir):
            if file.endswith(".json"):
                with open(os.path.join(job_dir, file), 'r') as f:
                    job = json.load(f)
                jobs.append(job)
    return jsonify(jobs)

@jobs_write_bp.route('/ingest', methods=['POST'])
def run_ingest():
    dataset = request.form.get('dataset')
    file = request.files.get('file')
    text_column = request.form.get('text_column')
    dataset_dir = os.path.join(DATA_DIR, dataset)
    if not os.path.exists(dataset_dir):
        os.makedirs(dataset_dir)
    file_path = os.path.join(dataset_dir, file.filename)
    file.save(file_path)

    job_id = str(uuid.uuid4())
    command = f'ls-ingest "{dataset}" --path="{file_path}"'
    if text_column:
        command += f' --text_column="{text_column}"'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/reingest', methods=['GET'])
def run_reingest():
    dataset = request.args.get('dataset')
    text_column = request.args.get('text_column')
    dataset_dir = os.path.join(DATA_DIR, dataset)
    file_path = os.path.join(dataset_dir, "input.parquet")

    job_id = str(uuid.uuid4())
    command = f'ls-ingest "{dataset}" --path="{file_path}"'
    if text_column:
        command += f' --text_column="{text_column}"'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})



@jobs_write_bp.route('/embed')
def run_embed():
    dataset = request.args.get('dataset')
    text_column = request.args.get('text_column')
    model_id = request.args.get('model_id') # model id
    prefix = request.args.get('prefix')
    dimensions = request.args.get('dimensions')
    batch_size = request.args.get('batch_size')
    max_seq_length = request.args.get('max_seq_length')

    job_id = str(uuid.uuid4())
    command = f'ls-embed "{dataset}" "{text_column}" "{model_id}" --prefix="{prefix}" --batch_size={batch_size}'
    if dimensions is not None:
        command += f" --dimensions={dimensions}"
    if max_seq_length is not None:
        command += f" --max_seq_length={max_seq_length}"
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/embed_truncate')
def run_embed_truncate():
    dataset = request.args.get('dataset')
    embedding_id = request.args.get('embedding_id') # model id
    dimensions = request.args.get('dimensions')

    job_id = str(uuid.uuid4())
    command = f'ls-embed-truncate "{dataset}" "{embedding_id}" {dimensions}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/embed_importer')
def run_embed_importer():
    dataset = request.args.get('dataset')
    model_id = request.args.get('model_id')
    embedding_column = request.args.get('embedding_column')
    text_column = request.args.get('text_column')

    job_id = str(uuid.uuid4())
    command = f'ls-embed-importer "{dataset}" "{embedding_column}" "{model_id}" "{text_column}"'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/rerun')
def rerun_job():
    dataset = request.args.get('dataset')
    job_id = request.args.get('job_id')
    # read the job file to get the command
    progress_file = os.path.join(DATA_DIR, dataset, "jobs", f"{job_id}.json")
    with open(progress_file, 'r') as f:
        job = json.load(f)
    command = job.get('command')
    command += f' --rerun {job.get("run_id")}'
    new_job_id = str(uuid.uuid4())
    print("new job id", new_job_id)
    threading.Thread(target=run_job, args=(dataset, new_job_id, command)).start()
    return jsonify({"job_id": new_job_id})

@jobs_write_bp.route('/kill')
def kill_job():
    dataset = request.args.get('dataset')
    job_id = request.args.get('job_id')
    # load the job file
    progress_file = os.path.join(DATA_DIR, dataset, "jobs", f"{job_id}.json")
    job = json.load(open(progress_file, 'r'))
    if job_id in PROCESSES:
        PROCESSES[job_id].kill()
        job["status"] = "dead"
        job["cause_of_death"] = "killed"
        with open(progress_file, 'w') as f:
            json.dump(job, f)
        return jsonify(job)
    else:
        job["status"] = "dead"
        job["cause_of_death"] = "process not found, presumed dead"
        with open(progress_file, 'w') as f:
            json.dump(job, f)
        return jsonify(job)

@jobs_write_bp.route('/delete/embedding')
def delete_embedding():
    dataset = request.args.get('dataset')
    embedding_id = request.args.get('embedding_id')

    # Get a list of all the umaps that have embedding_id in their .json so we can delete them too
    umap_dir = os.path.join(DATA_DIR, dataset, 'umaps')
    umaps_to_delete = []
    for file in os.listdir(umap_dir):
        if file.endswith(".json"):
            with open(os.path.join(umap_dir, file), 'r') as f:
                umap_data = json.load(f)
            if umap_data.get('embedding_id') == embedding_id:
                umaps_to_delete.append(file.replace('.json', ''))
    
    # Get a list of all the saes that have embedding_id in their .json so we can delete them too
    sae_dir = os.path.join(DATA_DIR, dataset, 'sae')
    if not os.path.exists(sae_dir):
        os.makedirs(sae_dir)
    saes_to_delete = []
    for file in os.listdir(sae_dir):
        if file.endswith(".json"):
            with open(os.path.join(sae_dir, file), 'r') as f:
                umap_data = json.load(f)
            if umap_data.get('embedding_id') == embedding_id:
                saes_to_delete.append(file.replace('.json', ''))


    job_id = str(uuid.uuid4())
    path = os.path.join(DATA_DIR, dataset, "embeddings", f"{embedding_id}*").replace(" ", "\\ ")
    command = f'rm -rf {path}'
    for umap in umaps_to_delete:
        delete_umap(dataset, umap)
    for sae in saes_to_delete:
        delete_sae(dataset, sae)
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/umap')
def run_umap():
    dataset = request.args.get('dataset')
    embedding_id = request.args.get('embedding_id')
    sae_id = request.args.get('sae_id')
    neighbors = request.args.get('neighbors')
    min_dist = request.args.get('min_dist')
    init = request.args.get('init')
    align = request.args.get('align')
    save = request.args.get('save')
    seed = request.args.get('seed')
    print("run umap", dataset, embedding_id, sae_id, neighbors, min_dist, init, align, save, seed)

    job_id = str(uuid.uuid4())
    command = f'ls-umap "{dataset}" "{embedding_id}" {neighbors} {min_dist}'
    if init:
        command += f' --init={init}'
    if align:
        command += f' --align={align}'
    if save:
        command += f' --save'
    if sae_id:
        command += f' --sae_id={sae_id}'
    if seed:
        command += f' --seed={seed}'

    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/delete/umap')
def delete_umap_request():
    dataset = request.args.get('dataset')
    umap_id = request.args.get('umap_id')
    return delete_umap(dataset, umap_id)

def delete_umap(dataset, umap_id):
    # Get a list of all the clusters that have umap_name in their .json so we can delete them too
    cluster_dir = os.path.join(DATA_DIR, dataset, 'clusters')
    clusters_to_delete = []
    for file in os.listdir(cluster_dir):
        if file.endswith(".json"):
            try:
                with open(os.path.join(cluster_dir, file), 'r') as f:
                    cluster_data = json.load(f)
                if cluster_data.get('umap_id') == umap_id:
                    clusters_to_delete.append(file.replace('.json', ''))
            except Exception as e:
                print("ERROR LOADING CLUSTER", file)
    

    job_id = str(uuid.uuid4())
    path = os.path.join(DATA_DIR, dataset, "umaps", f"{umap_id}*").replace(" ", "\\ ")
    command = f'rm -rf {path}'
    # Create the rm -rf commands from the clusters_to_delete list
    for cluster in clusters_to_delete:
        cpath = os.path.join(DATA_DIR, dataset, "clusters", f"{cluster}*").replace(" ", "\\ ")
        command += f'; rm -rf {cpath}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})



@jobs_write_bp.route('/sae')
def run_sae():
    dataset = request.args.get('dataset')
    embedding_id = request.args.get('embedding_id')
    model_id = request.args.get('model_id')
    k_expansion = request.args.get('k_expansion')

    job_id = str(uuid.uuid4())
    command = f'ls-sae "{dataset}" "{embedding_id}" "{model_id}" {k_expansion}'

    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/delete/sae')
def delete_sae_request():
    dataset = request.args.get('dataset')
    sae_id = request.args.get('sae_id')
    return delete_sae(dataset, sae_id)

def delete_sae(dataset, sae_id):
    # Get a list of all the umaps that have sae_id in their .json so we can delete them too
    umap_dir = os.path.join(DATA_DIR, dataset, 'umaps')
    umaps_to_delete = []
    for file in os.listdir(umap_dir):
        if file.endswith(".json"):
            with open(os.path.join(umap_dir, file), 'r') as f:
                umap_data = json.load(f)
            if umap_data.get('sae_id') == sae_id:
                umaps_to_delete.append(file.replace('.json', ''))
    

    job_id = str(uuid.uuid4())
    path = os.path.join(DATA_DIR, dataset, "saes", f"{sae_id}*").replace(" ", "\\ ")
    command = f'rm -rf {path}'
    for umap in umaps_to_delete:
        delete_umap(dataset, umap)
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/cluster')
def run_cluster():
    dataset = request.args.get('dataset')
    umap_id = request.args.get('umap_id')
    samples = request.args.get('samples')
    min_samples = request.args.get('min_samples')
    cluster_selection_epsilon = request.args.get('cluster_selection_epsilon')
    print("run cluster", dataset, umap_id, samples, min_samples, cluster_selection_epsilon)

    job_id = str(uuid.uuid4())
    command = f'ls-cluster "{dataset}" "{umap_id}" {samples} {min_samples} {cluster_selection_epsilon}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/delete/cluster')
def delete_cluster():
    dataset = request.args.get('dataset')
    cluster_id = request.args.get('cluster_id')
    job_id = str(uuid.uuid4())
    path = os.path.join(DATA_DIR, dataset, "clusters", f"{cluster_id}*").replace(" ", "\\ ")
    command = f'rm -rf {path}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/cluster_label')
def run_cluster_label():
    dataset = request.args.get('dataset')
    chat_id = request.args.get('chat_id')
    text_column = request.args.get('text_column')
    cluster_id = request.args.get('cluster_id')
    context = request.args.get('context')
    samples = request.args.get('samples')
    max_tokens_per_sample = request.args.get('max_tokens_per_sample')
    max_tokens_total = request.args.get('max_tokens_total')
    print("run cluster label", dataset, chat_id, text_column, cluster_id, samples, max_tokens_per_sample, max_tokens_total)
    if context:
        context = context.replace('"', '\\"')
    print("context", context)

    job_id = str(uuid.uuid4())
    command = f'ls-label "{dataset}" "{text_column}" "{cluster_id}" "{chat_id}" {samples} "{context}"'
    if max_tokens_per_sample:
        command += f' --max_tokens_per_sample={max_tokens_per_sample}'
    if max_tokens_total:
        command += f' --max_tokens_total={max_tokens_total}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/delete/cluster_label')
def delete_cluster_label():
    dataset = request.args.get('dataset')
    cluster_labels_id = request.args.get('cluster_labels_id')
    job_id = str(uuid.uuid4())
    path = os.path.join(DATA_DIR, dataset, "clusters", f"{cluster_labels_id}*").replace(" ", "\\ ")
    command = f'rm -rf {path}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})


@jobs_write_bp.route('/scope')
def run_scope():
    dataset = request.args.get('dataset')
    embedding_id = request.args.get('embedding_id')
    sae_id = request.args.get('sae_id')
    umap_id = request.args.get('umap_id')
    cluster_id = request.args.get('cluster_id')
    cluster_labels_id = request.args.get('cluster_labels_id')
    label = request.args.get('label')
    description = request.args.get('description')
    scope_id = request.args.get('scope_id')
    print("run scope", dataset, embedding_id, umap_id, cluster_id, cluster_labels_id, label, description, scope_id, sae_id)

    job_id = str(uuid.uuid4())
    command = f'ls-scope "{dataset}" "{embedding_id}" "{umap_id}" "{cluster_id}" "{cluster_labels_id}" "{label}" "{description}"'
    if sae_id:
        command += f' --sae_id={sae_id}'
    if scope_id:
        command += f' --scope_id={scope_id}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/delete/scope')
def delete_scope():
    dataset = request.args.get('dataset')
    scope_id = request.args.get('scope_id')

    job_id = str(uuid.uuid4())
    path = os.path.join(DATA_DIR, dataset, "scopes", f"{scope_id}*").replace(" ", "\\ ")
    command = f'rm -rf {path}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/plot')
def run_plot():
    dataset = request.args.get('dataset')
    scope_id = request.args.get('scope_id')
    config = request.args.get('config')
    print("run plot", dataset, scope_id, config)

    job_id = str(uuid.uuid4())
    command = f'ls-export-plot "{dataset}" {scope_id}'
    escaped_config = json.dumps(config)#.replace('"', '\\"')
    command += f' --plot_config={escaped_config}'
    print("command", command)

    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/download_dataset')
def download_dataset():
    dataset_repo = request.args.get('dataset_repo')
    dataset_name = request.args.get('dataset_name')

    job_id = str(uuid.uuid4())
    command = f'ls-download-dataset "{dataset_repo}" "{dataset_name}" "{DATA_DIR}"'
    threading.Thread(target=run_job, args=(dataset_name, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/upload_dataset')
def upload_dataset():
    dataset = request.args.get('dataset')
    hf_dataset = request.args.get('hf_dataset')
    main_parquet = request.args.get('main_parquet')
    private = request.args.get('private')

    job_id = str(uuid.uuid4())
    path = os.path.join(DATA_DIR, dataset)
    command = f'ls-upload-dataset "{path}" "{hf_dataset}" --main-parquet="{main_parquet}" --private={private}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})



import os
import re
import json
import argparse
from latentscope.util import get_data_dir
from latentscope import __version__


def main():
    print("MAIN")
    parser = argparse.ArgumentParser(description="Export scope as plot")
    parser.add_argument(
        "dataset_id", type=str, help="Dataset id (directory name in data folder)"
    )
    parser.add_argument("scope_id", type=str, help="Scope id")
    parser.add_argument(
        "--plot_config", type=str, default=None, help="Optional plot config"
    )

    args = parser.parse_args()
    print("ARGS", args)
    dmp(**vars(args))


def dmp(dataset_id, scope_id, plot_config=None):
    import datamapplot

    # import matplotlib
    # matplotlib.rcParams["figure.dpi"] = 300
    DATA_DIR = get_data_dir()
    print("DATA DIR", DATA_DIR)
    directory = os.path.join(DATA_DIR, dataset_id, "plots")
    if not os.path.exists(directory):
        os.makedirs(directory)

    def get_next_plot_number():
        # figure out the latest scope number
        plots_files = [
            f
            for f in os.listdir(directory)
            if re.match(rf"plots-{re.escape(scope_id)}-\d+\.(png|html)", f)
        ]
        if len(plots_files) > 0:
            last_plots = sorted(plots_files)[-1]
            last_plots_number = int(last_plots.split("-")[3].split(".")[0])
            next_plots_number = last_plots_number + 1
        else:
            next_plots_number = 1
        return next_plots_number

    next_plots_number = get_next_plot_number()
    id = f"plots-{scope_id}-{next_plots_number:03d}"

    print("RUNNING:", id)

    import pandas as pd

    plot = {
        "ls_version": __version__,
        "id": id,
        "scope_id": scope_id,
        "dataset_id": dataset_id,
    }

    # read each json file and add its contents to the scope file
    dataset_file = os.path.join(DATA_DIR, dataset_id, "meta.json")
    with open(dataset_file) as f:
        dataset = json.load(f)

    scope_file = os.path.join(DATA_DIR, dataset_id, "scopes", scope_id + ".json")
    with open(scope_file) as f:
        scope = json.load(f)

    print("loaded dataset and scope")
    # load the actual labels and save everything but the indices in a dict
    scope_parquet = pd.read_parquet(
        os.path.join(DATA_DIR, dataset_id, "scopes", scope_id + ".parquet")
    )
    print("loaded scope parquet", scope_parquet.columns)
    # remove the indices column
    import numpy as np

    xy_columns = scope_parquet[["x", "y"]].to_numpy()
    # set the label to "No topic" for anything with value -1 in raw_clusters in scope_parquet
    # TODO: we could make this optional, but part of the point of latent scope is that you assign things to clusters
    # scope_parquet.loc[scope_parquet['raw_cluster'] == -1, 'label'] = "No Topic"

    labels = scope_parquet["label"].to_numpy()
    label_sizes = pd.Series(labels).value_counts()
    label_sizes.reset_index()
    print(label_sizes)

    
    threshold = -1
    if plot_config:
        plot_config = json.loads(plot_config)
        threshold = plot_config.get("threshold", -1)
        del plot_config["threshold"]
        plot_config["noise_label"] = "No Topic"

    simplified_labels = scope_parquet["label"].copy()
    if threshold >= 0:
        simplified_labels[np.in1d(simplified_labels, label_sizes[label_sizes < threshold].index)] = "No Topic"

    highlight_labels = np.unique(simplified_labels)
    if plot_config is None:
        plot_config = {
            "label_over_points": True,
            "dynamic_label_size": True,
            "add_glow": True,
            "darkmode": False,
            "dpi": 150,
            "figsize": [24, 24],
            "label_wrap_width": 10,
            "point_size": 7,
            "max_font_size": 32,
            "min_font_size": 16,
            "min_font_weight": 100,
            "max_font_weight": 1000,
            "font_family": "Roboto Condensed",
            "glow_keywords": {
                "kernel_bandwidth": 0.01,
                "kernel": "exponential",
                "n_levels": 128,
                "max_alpha": 0.75,
            },
        }
    interactive_specific_options = {"point_radius_max_pixels": 4}
    is_interactive = "interactive" in plot_config and plot_config["interactive"]
    del plot_config["interactive"]
    print("making plot", xy_columns.shape, labels.shape)
    print("using config", plot_config)
    if is_interactive:
        print("making interactive plot")
        keys_to_delete = [
            "dpi",
            "label_over_points",
            "dynamic_label_size",
            "figsize",
            "point_size",
            "max_font_size",
            "min_font_size",
            "min_font_weight",
            "max_font_weight",
            "glow_keywords",
        ]
        for key in keys_to_delete:
            if key in plot_config:
                del plot_config[key]
        # add_glow
        del plot_config["add_glow"]
        plot = datamapplot.create_interactive_plot(
            xy_columns,
            simplified_labels,
            enable_search=True,
            use_medoids=True,
            point_radius_max_pixels=interactive_specific_options["point_radius_max_pixels"],
            **plot_config,
        )
        plot_path = os.path.join(directory, f"{id}.html")
        plot.save(plot_path)
        print(f"Plot saved to {plot_path}")
    else:
        fig, ax = datamapplot.create_plot(
            xy_columns,
            simplified_labels,
            force_matplotlib=True,
            use_medoids=True,
            highlight_labels=highlight_labels,
            **plot_config,
            # noise_label="No Topic",
            # label_wrap_width=10,
            # label_over_points=True,
            # dynamic_label_size=True,
            # dpi=150,
            # figsize=(24, 24),
            # point_size=7,
            # max_font_size=40,
            # min_font_size=16,
            # min_font_weight=100,
            # max_font_weight=1000,
            # font_family="Roboto Condensed",
            # add_glow=True,
            # glow_keywords={
            #     "kernel_bandwidth":0.01,
            #     "kernel":"exponential",
            #     "n_levels":128,
            #     "max_alpha":0.75
            # },
            # highlight_labels=highlight_labels,
            # darkmode=False
        )
        plot_path = os.path.join(directory, f"{id}.png")
        fig.savefig(plot_path)
        print(f"Plot saved to {plot_path}")

        config_path = os.path.join(directory, f"{id}.json")
        with open(config_path, "w", encoding="utf-8") as config_file:
            json.dump(plot_config, config_file, indent=4)

        print(f"Config saved to {config_path}")



"""
python latentscope/scripts/upload_scope.py ~/latent-scope-demo/datavis-misunderstood "ls-datavis-misunderstood" --main-parquet="scopes/scopes-001-input.parquet" --private=False
"""

from datasets import Dataset
from pathlib import Path
from huggingface_hub import login, HfApi
import os
import argparse
import shutil
import tempfile
from latentscope.util import get_key

def get_human_readable_size(size_in_bytes):
    """Convert bytes to human readable format."""
    for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
        if size_in_bytes < 1024.0:
            return f"{size_in_bytes:.1f} {unit}"
        size_in_bytes /= 1024.0
    return f"{size_in_bytes:.1f} PB"

def upload_to_huggingface(directory_path, dataset_name, main_parquet_path=None, token=None, private=True):
    """
    Upload a directory with all its files to Hugging Face datasets.
    
    Args:
        directory_path (str): Path to the directory to upload
        dataset_name (str): Name for the dataset on Hugging Face
        main_parquet_path (str): Path to the main parquet file relative to directory_path
        token (str, optional): Hugging Face API token
        private (bool, optional): Whether to make the dataset private (default: True)
    """
    # Get token from .env if not provided
    if not token:
        token = get_key("HUGGINGFACE_TOKEN")
        if token is None:
            raise ValueError("No token provided. Set HUGGINGFACE_TOKEN in .env or pass token argument")
    
    # Login to Hugging Face
    login(token)
    
    api = HfApi()
    # Get current user information
    user_info = api.whoami()
    username = user_info["name"]
    print("USERNAME", username)
    
    directory = Path(directory_path)

    # Create the repository if it doesn't exist
    try:
        print(f"Creating repository: {dataset_name}")
        repo_info = api.create_repo(
            repo_id=dataset_name,
            repo_type="dataset",
            private=private,
            exist_ok=True
        )
        print(f"Repository created/verified: {repo_info}")
    except Exception as e:
        print(f"Error creating repository: {e}")
        print(f"Token status: {'Token provided' if token else 'No token provided'}")
        print(f"Attempting to verify repository existence...")
        try:
            # Try to get the repository info to verify it exists and we have access
            repo_info = api.repo_info(repo_id=dataset_name, repo_type="dataset")
            print(f"Repository exists: {repo_info}")
        except Exception as verify_error:
            print(f"Could not verify repository: {verify_error}")
            raise

    print("DIRECTORY", directory)
    
    # Create a temporary directory for organizing files
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        
        # Create data directory for actual files
        data_dir = temp_path / "data"
        data_dir.mkdir(exist_ok=True)
        
        # Calculate total size before copying files
        total_size = 0
        for root, _, filenames in os.walk(directory):
            for filename in filenames:
                file_path = Path(root) / filename
                total_size += file_path.stat().st_size
        
        human_readable_size = get_human_readable_size(total_size)
        
        # Copy all files to the data directory while preserving structure
        for root, _, filenames in os.walk(directory):
            for filename in filenames:
                src_path = Path(root) / filename
                relative_path = src_path.relative_to(directory)
                dst_path = data_dir / relative_path
                
                # Create parent directories if they don't exist
                dst_path.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy2(src_path, dst_path)
        
        # If a main parquet file is specified, create a dataset from it
        if main_parquet_path:
            main_parquet = directory / main_parquet_path
            if not main_parquet.exists():
                raise FileNotFoundError(f"Main parquet file not found: {main_parquet}")
            
            print("MAIN PARQUET", main_parquet)
            # Load and push the dataset
            dataset = Dataset.from_parquet(str(main_parquet))
            dataset.push_to_hub(dataset_name, private=True)
        
        # Upload all files in the data directory
        api.upload_folder(
            repo_id=f"{username}/{dataset_name}",
            folder_path=str(data_dir),
            path_in_repo="latentscope",
            repo_type="dataset",
        )
        
        # Create a README if it doesn't exist
        readme_content = f"""
---
tags:
  - latent-scope
---
# {dataset_name}

This dataset contains the files necessary to view in [latentscope](https://github.com/enjalot/latent-scope).
The files in the `latentscope` are used by the app to view. You can also preview the scope TODO

Total size of dataset files: {human_readable_size}

TODO: download script inside latentscope
"""
        readme_path = temp_path / "README.md"
        with open(readme_path, "w") as f:
            f.write(readme_content)
        
        # Upload README
        api.upload_file(
            repo_id=f"{username}/{dataset_name}",
            path_or_fileobj=str(readme_path),
            path_in_repo="README.md",
            repo_type="dataset",
        )
        print(f"uploaded to: {username}/{dataset_name}")

def main():
    parser = argparse.ArgumentParser(description='Upload a directory with files to Hugging Face datasets')
    parser.add_argument('directory', help='Directory path to upload')
    parser.add_argument('dataset_name', help='Name for the dataset on Hugging Face (e.g., username/dataset-name)')
    parser.add_argument('--main-parquet', help='Path to main parquet file relative to directory', default=None)
    parser.add_argument('--private', help='Make the dataset private', default=True)
    parser.add_argument('--token', help='Hugging Face API token', default=None)
    
    args = parser.parse_args()
    print("ARGS", args)
    
    upload_to_huggingface(args.directory, args.dataset_name, args.main_parquet, args.token, args.private)

if __name__ == "__main__":
    main()


import os
import sys
import numpy as np
from flask import Blueprint, jsonify, request

# Create a Blueprint
tags_bp = Blueprint('tags_bp', __name__)
tags_write_bp = Blueprint('tags_write_bp', __name__)
DATA_DIR = os.getenv('LATENT_SCOPE_DATA')

# ===========================================================
# Tags
# ===========================================================

# cache of list of indices per tag per dataset
tagsets = {}

"""
Return the tagsets for a given dataset
This is a JSON object with the tag name as the key and an array of indices as the value
"""
@tags_bp.route("/", methods=['GET'])
def tags():
    dataset = request.args.get('dataset')
    tagdir = os.path.join(DATA_DIR, dataset, "tags")
    if not os.path.exists(tagdir):
        os.makedirs(tagdir)
    if dataset not in tagsets:
        tagsets[dataset] = {}
    # search the dataset directory for all files ending in .indices
    for f in os.listdir(tagdir):
        if f.endswith(".indices"):
            tag = f.split(".")[0]
            indices = np.loadtxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), dtype=int).tolist()
            if type(indices) == int:
                indices = [indices]
            tagsets[dataset][tag] = indices

    # return an object with the tags for a given dataset
    return jsonify(tagsets[dataset])

"""
Create a new tag for a given dataset
"""
@tags_write_bp.route("/new", methods=['GET'])
def new_tag():
    dataset = request.args.get('dataset')
    tag = request.args.get('tag')
    if dataset not in tagsets:
        tagsets[dataset] = {}
    # search the dataset directory for all files ending in .indices
    tags = []
    for f in os.listdir(os.path.join(DATA_DIR, dataset)):
        if f.endswith(".indices"):
            dtag = f.split(".")[0]
            indices = np.loadtxt(os.path.join(DATA_DIR, dataset, "tags", dtag + ".indices"), dtype=int).tolist()
            if type(indices) == int:
                indices = [indices]
            tagsets[dataset][dtag] = indices

    if tag not in tagsets[dataset]:
        tagsets[dataset][tag] = []
        # create an empty file
        filename = os.path.join(DATA_DIR, dataset, "tags", tag + ".indices")
        with open(filename, 'w') as f:
            f.write("")
            f.close()


    # return an object with the tags for a given dataset
    return jsonify(tagsets[dataset])

"""
Add a data index to a tag
"""
@tags_write_bp.route("/add", methods=['GET'])
def add_tag():
    dataset = request.args.get('dataset')
    tag = request.args.get('tag')
    index = request.args.get('index')
    if dataset not in tagsets:
        ts = tagsets[dataset] = {}
    else:
        ts = tagsets[dataset]
    if tag not in ts:
        # read a tag file, which is just a csv with a single column into an array of integers
        indices = np.loadtxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), dtype=int).tolist()
        if type(indices) == int:
            indices = [indices]
        ts[tag] = indices
    else:
        indices = ts[tag]

    if not indices:
        indices = []
    if index not in indices:
        indices.append(int(index))
        ts[tag] = indices
        # save the indices to a file
        np.savetxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), indices, fmt='%d')
    # return an object with the tags for a given dataset
    return jsonify(tagsets[dataset])

"""
Add data indices to a tag
"""
@tags_write_bp.route("/add", methods=['POST'])
def add_tags():
    data = request.get_json()
    dataset = data.get('dataset')
    tag = data.get('tag')
    new_indices = data.get('indices')
    # print("DATASET", dataset)
    # print("tag", tag)
    # print("indices", new_indices)

    if dataset not in tagsets:
        ts = tagsets[dataset] = {}
    else:
        ts = tagsets[dataset]
    if tag not in ts:
        # read a tag file, which is just a csv with a single column into an array of integers
        indices = np.loadtxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), dtype=int).tolist()
        if type(indices) == int:
            indices = [indices]
        ts[tag] = indices
    else:
        indices = ts[tag]

    if not indices:
        indices = []

    # new_indices_list = [int(idx) for idx in new_indices.split(',')]
    new_indices_list = [int(idx) for idx in new_indices]
    for idx in new_indices_list:
        if idx not in indices:
            indices.append(idx)
    ts[tag] = indices
    # save the indices to a file
    np.savetxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), indices, fmt='%d')
    # return an object with the tags for a given dataset
    return jsonify(tagsets[dataset])


"""
Remove a data index from a tag
"""
@tags_write_bp.route("/remove", methods=['GET'])
def remove_tag():
    dataset = request.args.get('dataset')
    tag = request.args.get('tag')
    index = int(request.args.get('index'))
    # print("dataset", dataset)
    # print("tag", tag)
    # print("index", index)
    if dataset not in tagsets:
        ts = tagsets[dataset] = {}
    else:
        ts = tagsets[dataset]
    if tag not in ts:
        # read a tag file, which is just a csv with a single column into an array of integers
        indices = np.loadtxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), dtype=int).tolist()
        if type(indices) == int:
            indices = [indices]
        ts[tag] = indices
    else:
        indices = ts[tag]
    print("indices", indices)
    if index in indices:
        print("Removing", index)
        indices.remove(index)
        print("removed", indices)
        ts[tag] = indices
        # save the indices to a file
        np.savetxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), indices, fmt='%d')
    print("returning", tagsets[dataset])
    # return an object with the tags for a given dataset
    return jsonify(tagsets[dataset])


"""
Add data indices to a tag
"""
@tags_write_bp.route("/remove", methods=['POST'])
def remove_tags():
    data = request.get_json()
    dataset = data.get('dataset')
    tag = data.get('tag')
    remove_indices = data.get('indices')
    print("dataset", dataset)
    print("tag", tag)
    print("indices", remove_indices)

    if dataset not in tagsets:
        ts = tagsets[dataset] = {}
    else:
        ts = tagsets[dataset]
    if tag not in ts:
        # read a tag file, which is just a csv with a single column into an array of integers
        indices = np.loadtxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), dtype=int).tolist()
        if type(indices) == int:
            indices = [indices]
        ts[tag] = indices
    else:
        indices = ts[tag]

    if not indices:
        indices = []

    # new_indices_list = [int(idx) for idx in new_indices.split(',')]
    new_indices_list = [int(idx) for idx in remove_indices]
    for idx in new_indices_list:
        if idx in indices:
            indices.remove(idx)
    ts[tag] = indices
    # save the indices to a file
    np.savetxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), indices, fmt='%d')
    # return an object with the tags for a given dataset
    return jsonify(tagsets[dataset])


@tags_write_bp.route("/delete", methods=['GET'])
def delete_tag():
    dataset = request.args.get('dataset')
    tag = request.args.get('tag')
    if dataset not in tagsets:
        ts = tagsets[dataset] = {}
    else:
        ts = tagsets[dataset]
    if tag in ts:
        del ts[tag]
    try:
        os.remove(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"))
    except FileNotFoundError:
        pass
    return jsonify(tagsets[dataset])




import os
import re
import json
import fnmatch
import pandas as pd
from flask import Blueprint, jsonify, request

# Create a Blueprint
datasets_bp = Blueprint('datasets_bp', __name__)
datasets_write_bp = Blueprint('datasets_write_bp', __name__)
DATA_DIR = os.getenv('LATENT_SCOPE_DATA')

"""
Get the essential metadata for all available datasets.
Essential metadata is stored in meta.json
"""
@datasets_bp.route('/', methods=['GET'])
def get_datasets():
    datasets = []

    for dir in os.listdir(DATA_DIR):
        file_path = os.path.join(DATA_DIR, dir, 'meta.json')
        if os.path.isfile(file_path):
            with open(file_path, 'r', encoding='utf-8') as file:
                try:
                    jsonData = json.load(file)
                    jsonData['id'] = dir
                    datasets.append(jsonData)
                except:
                    print("Error reading meta.json", file_path)

    datasets.sort(key=lambda x: x.get('id'))
    return jsonify(datasets)

"""
Get all metadata files from the given a directory.
"""
def scan_for_json_files(directory_path, match_pattern=r".*\.json$"):
    try:
        # files = os.listdir(directory_path)
        files = sorted(os.listdir(directory_path), key=lambda x: os.path.getmtime(os.path.join(directory_path, x)), reverse=True)
    except OSError as err:
        print('Unable to scan directory:', err)
        return jsonify({"error": "Unable to scan directory"}), 500

    json_files = [file for file in files if re.match(match_pattern, file)]
    # print("files", files)
    # print("json", json_files)

    json_contents = []
    for file in json_files:
        try:
            with open(os.path.join(directory_path, file), 'r', encoding='utf-8') as json_file:
                json_contents.append(json.load(json_file))
        except json.JSONDecodeError as err:
            print('Error parsing JSON string:', err)
    return jsonify(json_contents)

@datasets_bp.route('/<dataset>/meta', methods=['GET'])
def get_dataset_meta(dataset):
    file_path = os.path.join(DATA_DIR, dataset, "meta.json")
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)
    return jsonify(json_contents)

@datasets_write_bp.route('/<dataset>/meta/update', methods=['GET'])
def update_dataset_meta(dataset):
    key = request.args.get('key')
    value = request.args.get('value')
    try:
        value = json.loads(value)
    except json.JSONDecodeError as err:
        print("Invalid JSON format for value", value, err)

    file_path = os.path.join(DATA_DIR, dataset, "meta.json")
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)
    json_contents[key] = value
    # write the file back out
    with open(file_path, 'w', encoding='utf-8') as json_file:
        json.dump(json_contents, json_file)
    return jsonify(json_contents)

@datasets_bp.route('/<dataset>/embeddings', methods=['GET'])
def get_dataset_embeddings(dataset):
    directory_path = os.path.join(DATA_DIR, dataset, "embeddings")
    # directory_path = os.path.join(DATA_DIR, dataset, "umaps")
    return scan_for_json_files(directory_path)

@datasets_bp.route('/<dataset>/embeddings/<embedding>', methods=['GET'])
def get_dataset_embedding(dataset, embedding):
    file_path = os.path.join(DATA_DIR, dataset, "embeddings", embedding + ".json")
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)
    return jsonify(json_contents)

@datasets_bp.route('/<dataset>/saes', methods=['GET'])
def get_dataset_saes(dataset):
    directory_path = os.path.join(DATA_DIR, dataset, "saes")
    # directory_path = os.path.join(DATA_DIR, dataset, "umaps")
    return scan_for_json_files(directory_path)

@datasets_bp.route('/<dataset>/saes/<sae>', methods=['GET'])
def get_dataset_sae(dataset, sae):
    file_path = os.path.join(DATA_DIR, dataset, "saes", sae + ".json")
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)
    return jsonify(json_contents)

@datasets_bp.route('/<dataset>/umaps', methods=['GET'])
def get_dataset_umaps(dataset):
    directory_path = os.path.join(DATA_DIR, dataset, "umaps")
    return scan_for_json_files(directory_path)

@datasets_bp.route('/<dataset>/umaps/<umap>', methods=['GET'])
def get_dataset_umap(dataset, umap):
    file_path = os.path.join(DATA_DIR, dataset, "umaps", umap + ".json")
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)
    return jsonify(json_contents)

@datasets_bp.route('/<dataset>/umaps/<umap>/points', methods=['GET'])
def get_dataset_umap_points(dataset, umap):
    file_path = os.path.join(DATA_DIR, dataset, "umaps", umap + ".parquet")
    df = pd.read_parquet(file_path)
    return df.to_json(orient="records")

@datasets_bp.route('/<dataset>/clusters', methods=['GET'])
def get_dataset_clusters(dataset):
    directory_path = os.path.join(DATA_DIR, dataset, "clusters")
    return scan_for_json_files(directory_path, match_pattern=r"cluster-\d+\.json")

@datasets_bp.route('/<dataset>/clusters/<cluster>', methods=['GET'])
def get_dataset_cluster(dataset, cluster):
    file_path = os.path.join(DATA_DIR, dataset, "clusters", cluster + ".json")
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)
    return jsonify(json_contents)

# @datasets_bp.route('/<dataset>/clusters/<cluster>/labels', methods=['GET'])
# def get_dataset_cluster_labels_default(dataset, cluster):
#     file_name = cluster + "-labels.parquet"
#     file_path = os.path.join(DATA_DIR, dataset, "clusters", file_name)
#     df = pd.read_parquet(file_path)
#     return df.to_json(orient="records")

@datasets_bp.route('/<dataset>/clusters/<cluster>/indices', methods=['GET'])
def get_dataset_cluster_indices(dataset, cluster):
    file_name = cluster + ".parquet"
    file_path = os.path.join(DATA_DIR, dataset, "clusters", file_name)
    df = pd.read_parquet(file_path)
    return df.to_json(orient="records")

@datasets_bp.route('/<dataset>/clusters/<cluster>/labels/<id>', methods=['GET'])
def get_dataset_cluster_labels(dataset, cluster, id):
    # if model == "default":
    #     return get_dataset_cluster_labels_default(dataset, cluster)
    file_name = cluster + "-labels-" + id + ".parquet"
    file_path = os.path.join(DATA_DIR, dataset, "clusters", file_name)
    df = pd.read_parquet(file_path)
    df.reset_index(inplace=True)
    return df.to_json(orient="records")

# This was rewritten in bulk.py to only affect a scope
# @datasets_write_bp.route('/<dataset>/clusters/<cluster>/labels/<id>/label/<index>', methods=['GET'])
# def overwrite_dataset_cluster_label(dataset, cluster, id, index):
#     index = int(index)
#     new_label = request.args.get('label')
#     print("write label", index, new_label)
#     if new_label is None:
#         return jsonify({"error": "Missing 'label' in request data"}), 400

#     file_name = cluster + "-labels-" + id + ".parquet"
#     file_path = os.path.join(DATA_DIR, dataset, "clusters", file_name)
#     try:
#         df = pd.read_parquet(file_path)
#     except FileNotFoundError:
#         return jsonify({"error": "File not found"}), 404

#     if index >= len(df):
#         return jsonify({"error": "Index out of range"}), 400

#     df.at[index, 'label'] = new_label
#     df.to_parquet(file_path)

#     return jsonify({"success": True, "message": "Label updated successfully"})


@datasets_bp.route('/<dataset>/clusters/<cluster>/labels_available', methods=['GET'])
def get_dataset_cluster_labels_available(dataset, cluster):
    directory_path = os.path.join(DATA_DIR, dataset, "clusters")
    return scan_for_json_files(directory_path, match_pattern=rf"{cluster}-labels-.*\.json")
    # try:
    #     files = sorted(os.listdir(directory_path), key=lambda x: os.path.getmtime(os.path.join(directory_path, x)), reverse=True)
    # except OSError as err:
    #     print('Unable to scan directory:', err)
    #     return jsonify({"error": "Unable to scan directory"}), 500

    # pattern = re.compile(r'^' + cluster + '-labels-(.*).parquet$')
    # model_names = [pattern.match(file).group(1) for file in files if pattern.match(file)]
    # return jsonify(model_names)


def get_next_scopes_number(dataset):
    # figure out the latest scope number
    scopes_files = [f for f in os.listdir(os.path.join(DATA_DIR,dataset,"scopes")) if re.match(r"scopes-\d+\.json", f)]
    if len(scopes_files) > 0:
        last_scopes = sorted(scopes_files)[-1]
        last_scopes_number = int(last_scopes.split("-")[1].split(".")[0])
        next_scopes_number = last_scopes_number + 1
    else:
        next_scopes_number = 1
    return next_scopes_number

@datasets_bp.route('/<dataset>/scopes', methods=['GET'])
def get_dataset_scopes(dataset):
    directory_path = os.path.join(DATA_DIR, dataset, "scopes")
    print("dataset", dataset, directory_path)
    return scan_for_json_files(directory_path, match_pattern=r".*[0-9]+\.json$")

@datasets_bp.route('/<dataset>/scopes/<scope>', methods=['GET'])
def get_dataset_scope(dataset, scope):
    directory_path = os.path.join(DATA_DIR, dataset, "scopes")
    file_path = os.path.join(directory_path, scope + ".json")
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)
    return jsonify(json_contents)

@datasets_bp.route('/<dataset>/scopes/<scope>/parquet', methods=['GET'])
def get_dataset_scope_parquet(dataset, scope):
    directory_path = os.path.join(DATA_DIR, dataset, "scopes")
    file_path = os.path.join(directory_path, scope + ".parquet")
    df = pd.read_parquet(file_path)
    return df.to_json(orient="records")

@datasets_write_bp.route('/<dataset>/scopes/<scope>/description', methods=['GET'])
def overwrite_scope_description(dataset, scope):
    new_label = request.args.get('label')
    new_description = request.args.get('description')

    file_name = scope + ".json"
    file_path = os.path.join(DATA_DIR, dataset, "scopes", file_name)
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)

    json_contents['label'] = new_label
    json_contents['description'] = new_description

    with open(file_path, 'w', encoding='utf-8') as json_file:
        json.dump(json_contents, json_file)
    
    return jsonify({"success": True, "message": "Description updated successfully"})

@datasets_write_bp.route('/<dataset>/scopes/<scope>/new-cluster', methods=['GET'])
def new_scope_cluster(dataset, scope):
    new_label = request.args.get('label')

    file_name = scope + ".json"
    file_path = os.path.join(DATA_DIR, dataset, "scopes", file_name)
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)

    clusters = json_contents.get('cluster_labels_lookup', [])
    clusterIndex = len(clusters)
    clusters.append({
        "cluster": clusterIndex, 
        "label": new_label,
        "hull": [],
        "description": ""
    })
    json_contents['cluster_labels_lookup'] = clusters

    with open(file_path, 'w', encoding='utf-8') as json_file:
        json.dump(json_contents, json_file)
    
    return jsonify({"success": True, "message": "Description updated successfully"})


@datasets_bp.route('/<dataset>/export/list', methods=['GET'])
def get_dataset_export_list(dataset):
    directory_path = os.path.join(DATA_DIR, dataset)
    print("dataset", dataset, directory_path)
    # scan the directory for files and directories
    # then walk the directories to find all the files
    # then return the list of files
    file_list = []
    for root, dirs, files in os.walk(directory_path):
        if "jobs" in root:
            continue
        for file in files:
            if file == ".DS_Store":
                continue
            full_path = os.path.join(root, file)
            file_name = os.path.basename(full_path)
            relative_path = os.path.relpath(full_path, directory_path)
            directory = os.path.relpath(root, directory_path)
            size = os.path.getsize(full_path)
            file_list.append((file_name, directory, relative_path, full_path, size))

    return jsonify(file_list)

@datasets_bp.route('/<dataset>/plot/<scope>/list', methods=['GET'])
def get_dataset_plot_list(dataset, scope):
    directory_path = os.path.join(DATA_DIR, dataset, "plots")
    print("dataset", dataset, directory_path)
    if not os.path.exists(directory_path):
        return jsonify([])
    # scan the directory for files and directories
    # then walk the directories to find all the files
    # then return the list of files
    file_list = []
    files = [f for f in os.listdir(directory_path) if os.path.isfile(os.path.join(directory_path, f))]
    for file in files:
        if not (file.endswith(".png") and scope in file):
            continue
        full_path = os.path.join(directory_path, file)
        file_name = os.path.basename(full_path)
        size = os.path.getsize(full_path)
        file_list.append((file_name, full_path, size))

    return jsonify(file_list)


"""
Usage:
python latentscope/scripts/download_dataset.py "enjalot/ls-datavis-misunderstood" ~/latent-scope-data/datavis-misunderstood
"""

from datasets import load_dataset
from huggingface_hub import hf_hub_download, snapshot_download
from pathlib import Path
import argparse
import os
from latentscope.util import get_key

def download_from_huggingface(dataset_repo, dataset_name,output_dir,token=None):
    """
    Download a latentscope dataset from Hugging Face.
    
    Args:
        dataset_path (str): Path to the dataset on Hugging Face (e.g., 'username/dataset-name')
        output_dir (str): Local directory to save the downloaded files
        token (str, optional): Hugging Face API token
    """
    # Get token from .env if not provided
    if not token:
        token = get_key("HUGGINGFACE_TOKEN")
    
    output_path = Path(output_dir)
    output_path.mkdir(parents=True, exist_ok=True)
    
    try:
        # Download the latentscope directory into output_dir/dataset_name
        latentscope_path = Path(output_dir) / dataset_name
        # if the directory already exists, raise an error
        if latentscope_path.exists():
            # raise Exception(f"Directory {latentscope_path} already exists")
            print(f"Warning: directory {latentscope_path} already exists")
        latentscope_path.mkdir(parents=True, exist_ok=True)

        # Download all files from the latentscope directory
        snapshot_download(
            repo_id=dataset_repo,
            repo_type="dataset",
            local_dir=str(latentscope_path),
            token=token,
            local_dir_use_symlinks=False
        )

        # Delete data and readme directories if they exist
        data_path = latentscope_path / "data"
        readme_path = latentscope_path / "README.md"
        if data_path.exists():
            import shutil
            shutil.rmtree(data_path)
        if readme_path.exists():
            os.remove(readme_path)
        # Move contents of latentscope directory up one level
        latentscope_dir = latentscope_path / "latentscope"
        if latentscope_dir.exists():
            import shutil
            # Move all contents up one level
            for item in latentscope_dir.iterdir():
                dest = latentscope_path / item.name
                if dest.exists():
                    if dest.is_dir():
                        shutil.rmtree(dest)
                    else:
                        os.remove(dest)
                shutil.move(str(item), str(dest))
            # Remove the now empty latentscope directory
            latentscope_dir.rmdir()

        # Update meta.json with new dataset_name
        meta_path = latentscope_path / "meta.json"
        if meta_path.exists():
            import json
            with open(meta_path, 'r') as f:
                meta = json.load(f)
            meta['id'] = dataset_name
            with open(meta_path, 'w') as f:
                json.dump(meta, f, indent=2)

        print(f"Successfully downloaded latentscope files to: {latentscope_path}")
            
    except Exception as e:
        print(f"Error downloading scope: {e}")
        raise

def main():
    parser = argparse.ArgumentParser(description='Download a latentscope dataset from Hugging Face')
    parser.add_argument('dataset_repo', help='Path to the dataset on Hugging Face (e.g., username/dataset-name)')
    parser.add_argument('dataset_name', help='Name of the dataset')
    parser.add_argument('output_dir', help='Local directory to save the downloaded files')
    parser.add_argument('--token', help='Hugging Face API token', default=None)
    
    args = parser.parse_args()
    
    download_from_huggingface(args.dataset_repo, args.dataset_name, args.output_dir, args.token)

if __name__ == "__main__":
    main()



import os
import re
import json
import fnmatch
import pandas as pd
from flask import Blueprint, jsonify, request

# Create a Blueprint
admin_bp = Blueprint('admin_bp', __name__)
DATA_DIR = os.getenv('LATENT_SCOPE_DATA')

@admin_bp.route('/', methods=['GET'])
def get_datasets():
    datasets = []

    for dir in os.listdir(DATA_DIR):
        scopes_dir = os.path.join(DATA_DIR, dir, 'scopes')

        meta_file_path = os.path.join(DATA_DIR, dir, 'meta.json')
        entry = {'id': dir}
        if os.path.exists(meta_file_path):
            with open(meta_file_path, 'r', encoding='utf-8') as meta_file:
                try:
                    meta_contents = json.load(meta_file)
                    length = meta_contents.get('length', None)
                    entry['length'] = length
                except json.JSONDecodeError as e:
                    print(f"Error reading meta.json for {dir}: {e}")

        if os.path.isdir(scopes_dir):
            scope_files = [f for f in os.listdir(scopes_dir) if f.startswith('scopes-') and f.endswith('.parquet') and 'input' not in f]
            if scope_files:
                dataset_scopes = []
                for scope_file in scope_files:
                    dataset_scopes.append(scope_file.replace(".parquet", ""))
                entry['scopes'] = dataset_scopes
                datasets.append(entry)

    # Convert the datasets list into a nested HTML list
    html_list = "<ul>"
    for dataset in datasets:
        html_list += f'<li style="margin-left: 10px; margin-bottom: 10px;">{dataset["id"]} ({dataset.get("length", "unknown")} rows)'
        if dataset['scopes']:
            html_list += "<ul>"
            for scope in dataset['scopes']:
                html_list += f'<li><a href="/api/admin/dataset/{dataset["id"]}/scope/{scope}">{scope}</a></li>'
            html_list += "</ul>"
        html_list += "</li>"
    html_list += "</ul>"
    datasets = html_list

    return html_list



@admin_bp.route('/dataset/<dataset>/scope/<scope>', methods=['GET'])
def get_dataset_scope(dataset, scope):

    # Read the input parquet file for the dataset
    input_df = pd.read_parquet(os.path.join(DATA_DIR, dataset, "input.parquet"))

    # Read the scope parquet file for the dataset
    scope_df = pd.read_parquet(os.path.join(DATA_DIR, dataset, "scopes", scope + ".parquet"))

    # Convert both dataframes to HTML tables with styling
    input_html = input_df.to_html(
        classes='table table-striped',
        border=0,
        index=True,
        escape=False,
        max_rows=100
    )
    
    scope_html = scope_df.to_html(
        classes='table table-striped', 
        border=0,
        index=True,
        escape=False,
        max_rows=100
    )

    # Combine the tables with headers
    html = f"""
    <h3>{dataset}</h3>
    {input_html}
    <h3>{scope}</h3>
    {scope_html}
    """

    return html



# @datasets_bp.route('/<dataset>/meta', methods=['GET'])
# def get_dataset_meta(dataset):
#     file_path = os.path.join(DATA_DIR, dataset, "meta.json")
#     with open(file_path, 'r', encoding='utf-8') as json_file:
#         json_contents = json.load(json_file)
#     return jsonify(json_contents)

# @datasets_write_bp.route('/<dataset>/meta/update', methods=['GET'])
# def update_dataset_meta(dataset):
#     key = request.args.get('key')
#     value = request.args.get('value')
#     try:
#         value = json.loads(value)
#     except json.JSONDecodeError as err:
#         print("Invalid JSON format for value", value, err)

#     file_path = os.path.join(DATA_DIR, dataset, "meta.json")
#     with open(file_path, 'r', encoding='utf-8') as json_file:
#         json_contents = json.load(json_file)
#     json_contents[key] = value
#     # write the file back out
#     with open(file_path, 'w', encoding='utf-8') as json_file:
#         json.dump(json_contents, json_file)
#     return jsonify(json_contents)

# @datasets_bp.route('/<dataset>/embeddings', methods=['GET'])
# def get_dataset_embeddings(dataset):
#     directory_path = os.path.join(DATA_DIR, dataset, "embeddings")
#     # directory_path = os.path.join(DATA_DIR, dataset, "umaps")
#     return scan_for_json_files(directory_path)

# @datasets_bp.route('/<dataset>/embeddings/<embedding>', methods=['GET'])
# def get_dataset_embedding(dataset, embedding):
#     file_path = os.path.join(DATA_DIR, dataset, "embeddings", embedding + ".json")
#     with open(file_path, 'r', encoding='utf-8') as json_file:
#         json_contents = json.load(json_file)
#     return jsonify(json_contents)

# @datasets_bp.route('/<dataset>/saes', methods=['GET'])
# def get_dataset_saes(dataset):
#     directory_path = os.path.join(DATA_DIR, dataset, "saes")
#     # directory_path = os.path.join(DATA_DIR, dataset, "umaps")
#     return scan_for_json_files(directory_path)

# @datasets_bp.route('/<dataset>/saes/<sae>', methods=['GET'])
# def get_dataset_sae(dataset, sae):
#     file_path = os.path.join(DATA_DIR, dataset, "saes", sae + ".json")
#     with open(file_path, 'r', encoding='utf-8') as json_file:
#         json_contents = json.load(json_file)
#     return jsonify(json_contents)

# @datasets_bp.route('/<dataset>/umaps', methods=['GET'])
# def get_dataset_umaps(dataset):
#     directory_path = os.path.join(DATA_DIR, dataset, "umaps")
#     return scan_for_json_files(directory_path)

# @datasets_bp.route('/<dataset>/umaps/<umap>', methods=['GET'])
# def get_dataset_umap(dataset, umap):
#     file_path = os.path.join(DATA_DIR, dataset, "umaps", umap + ".json")
#     with open(file_path, 'r', encoding='utf-8') as json_file:
#         json_contents = json.load(json_file)
#     return jsonify(json_contents)

# @datasets_bp.route('/<dataset>/umaps/<umap>/points', methods=['GET'])
# def get_dataset_umap_points(dataset, umap):
#     file_path = os.path.join(DATA_DIR, dataset, "umaps", umap + ".parquet")
#     df = pd.read_parquet(file_path)
#     return df.to_json(orient="records")

# @datasets_bp.route('/<dataset>/clusters', methods=['GET'])
# def get_dataset_clusters(dataset):
#     directory_path = os.path.join(DATA_DIR, dataset, "clusters")
#     return scan_for_json_files(directory_path, match_pattern=r"cluster-\d+\.json")

# @datasets_bp.route('/<dataset>/clusters/<cluster>', methods=['GET'])
# def get_dataset_cluster(dataset, cluster):
#     file_path = os.path.join(DATA_DIR, dataset, "clusters", cluster + ".json")
#     with open(file_path, 'r', encoding='utf-8') as json_file:
#         json_contents = json.load(json_file)
#     return jsonify(json_contents)

# # @datasets_bp.route('/<dataset>/clusters/<cluster>/labels', methods=['GET'])
# # def get_dataset_cluster_labels_default(dataset, cluster):
# #     file_name = cluster + "-labels.parquet"
# #     file_path = os.path.join(DATA_DIR, dataset, "clusters", file_name)
# #     df = pd.read_parquet(file_path)
# #     return df.to_json(orient="records")

# @datasets_bp.route('/<dataset>/clusters/<cluster>/indices', methods=['GET'])
# def get_dataset_cluster_indices(dataset, cluster):
#     file_name = cluster + ".parquet"
#     file_path = os.path.join(DATA_DIR, dataset, "clusters", file_name)
#     df = pd.read_parquet(file_path)
#     return df.to_json(orient="records")

# @datasets_bp.route('/<dataset>/clusters/<cluster>/labels/<id>', methods=['GET'])
# def get_dataset_cluster_labels(dataset, cluster, id):
#     # if model == "default":
#     #     return get_dataset_cluster_labels_default(dataset, cluster)
#     file_name = cluster + "-labels-" + id + ".parquet"
#     file_path = os.path.join(DATA_DIR, dataset, "clusters", file_name)
#     df = pd.read_parquet(file_path)
#     df.reset_index(inplace=True)
#     return df.to_json(orient="records")

# # This was rewritten in bulk.py to only affect a scope
# # @datasets_write_bp.route('/<dataset>/clusters/<cluster>/labels/<id>/label/<index>', methods=['GET'])
# # def overwrite_dataset_cluster_label(dataset, cluster, id, index):
# #     index = int(index)
# #     new_label = request.args.get('label')
# #     print("write label", index, new_label)
# #     if new_label is None:
# #         return jsonify({"error": "Missing 'label' in request data"}), 400

# #     file_name = cluster + "-labels-" + id + ".parquet"
# #     file_path = os.path.join(DATA_DIR, dataset, "clusters", file_name)
# #     try:
# #         df = pd.read_parquet(file_path)
# #     except FileNotFoundError:
# #         return jsonify({"error": "File not found"}), 404

# #     if index >= len(df):
# #         return jsonify({"error": "Index out of range"}), 400

# #     df.at[index, 'label'] = new_label
# #     df.to_parquet(file_path)

# #     return jsonify({"success": True, "message": "Label updated successfully"})


# @datasets_bp.route('/<dataset>/clusters/<cluster>/labels_available', methods=['GET'])
# def get_dataset_cluster_labels_available(dataset, cluster):
#     directory_path = os.path.join(DATA_DIR, dataset, "clusters")
#     return scan_for_json_files(directory_path, match_pattern=rf"{cluster}-labels-.*\.json")
#     # try:
#     #     files = sorted(os.listdir(directory_path), key=lambda x: os.path.getmtime(os.path.join(directory_path, x)), reverse=True)
#     # except OSError as err:
#     #     print('Unable to scan directory:', err)
#     #     return jsonify({"error": "Unable to scan directory"}), 500

#     # pattern = re.compile(r'^' + cluster + '-labels-(.*).parquet$')
#     # model_names = [pattern.match(file).group(1) for file in files if pattern.match(file)]
#     # return jsonify(model_names)


# def get_next_scopes_number(dataset):
#     # figure out the latest scope number
#     scopes_files = [f for f in os.listdir(os.path.join(DATA_DIR,dataset,"scopes")) if re.match(r"scopes-\d+\.json", f)]
#     if len(scopes_files) > 0:
#         last_scopes = sorted(scopes_files)[-1]
#         last_scopes_number = int(last_scopes.split("-")[1].split(".")[0])
#         next_scopes_number = last_scopes_number + 1
#     else:
#         next_scopes_number = 1
#     return next_scopes_number

# @datasets_bp.route('/<dataset>/scopes', methods=['GET'])
# def get_dataset_scopes(dataset):
#     directory_path = os.path.join(DATA_DIR, dataset, "scopes")
#     print("dataset", dataset, directory_path)
#     return scan_for_json_files(directory_path, match_pattern=r".*[0-9]+\.json$")

# @datasets_bp.route('/<dataset>/scopes/<scope>', methods=['GET'])
# def get_dataset_scope(dataset, scope):
#     directory_path = os.path.join(DATA_DIR, dataset, "scopes")
#     file_path = os.path.join(directory_path, scope + ".json")
#     with open(file_path, 'r', encoding='utf-8') as json_file:
#         json_contents = json.load(json_file)
#     return jsonify(json_contents)

# @datasets_bp.route('/<dataset>/scopes/<scope>/parquet', methods=['GET'])
# def get_dataset_scope_parquet(dataset, scope):
#     directory_path = os.path.join(DATA_DIR, dataset, "scopes")
#     file_path = os.path.join(directory_path, scope + ".parquet")
#     df = pd.read_parquet(file_path)
#     return df.to_json(orient="records")

# @datasets_write_bp.route('/<dataset>/scopes/<scope>/description', methods=['GET'])
# def overwrite_scope_description(dataset, scope):
#     new_label = request.args.get('label')
#     new_description = request.args.get('description')

#     file_name = scope + ".json"
#     file_path = os.path.join(DATA_DIR, dataset, "scopes", file_name)
#     with open(file_path, 'r', encoding='utf-8') as json_file:
#         json_contents = json.load(json_file)

#     json_contents['label'] = new_label
#     json_contents['description'] = new_description

#     with open(file_path, 'w', encoding='utf-8') as json_file:
#         json.dump(json_contents, json_file)
    
#     return jsonify({"success": True, "message": "Description updated successfully"})

# @datasets_write_bp.route('/<dataset>/scopes/<scope>/new-cluster', methods=['GET'])
# def new_scope_cluster(dataset, scope):
#     new_label = request.args.get('label')

#     file_name = scope + ".json"
#     file_path = os.path.join(DATA_DIR, dataset, "scopes", file_name)
#     with open(file_path, 'r', encoding='utf-8') as json_file:
#         json_contents = json.load(json_file)

#     clusters = json_contents.get('cluster_labels_lookup', [])
#     clusterIndex = len(clusters)
#     clusters.append({
#         "cluster": clusterIndex, 
#         "label": new_label,
#         "hull": [],
#         "description": ""
#     })
#     json_contents['cluster_labels_lookup'] = clusters

#     with open(file_path, 'w', encoding='utf-8') as json_file:
#         json.dump(json_contents, json_file)
    
#     return jsonify({"success": True, "message": "Description updated successfully"})


# @datasets_bp.route('/<dataset>/export/list', methods=['GET'])
# def get_dataset_export_list(dataset):
#     directory_path = os.path.join(DATA_DIR, dataset)
#     print("dataset", dataset, directory_path)
#     # scan the directory for files and directories
#     # then walk the directories to find all the files
#     # then return the list of files
#     file_list = []
#     for root, dirs, files in os.walk(directory_path):
#         if "jobs" in root:
#             continue
#         for file in files:
#             if file == ".DS_Store":
#                 continue
#             full_path = os.path.join(root, file)
#             file_name = os.path.basename(full_path)
#             relative_path = os.path.relpath(full_path, directory_path)
#             directory = os.path.relpath(root, directory_path)
#             size = os.path.getsize(full_path)
#             file_list.append((file_name, directory, relative_path, full_path, size))

#     return jsonify(file_list)

# @datasets_bp.route('/<dataset>/plot/<scope>/list', methods=['GET'])
# def get_dataset_plot_list(dataset, scope):
#     directory_path = os.path.join(DATA_DIR, dataset, "plots")
#     print("dataset", dataset, directory_path)
#     # scan the directory for files and directories
#     # then walk the directories to find all the files
#     # then return the list of files
#     file_list = []
#     files = [f for f in os.listdir(directory_path) if os.path.isfile(os.path.join(directory_path, f))]
#     for file in files:
#         if not (file.endswith(".png") and scope in file):
#             continue
#         full_path = os.path.join(directory_path, file)
#         file_name = os.path.basename(full_path)
#         size = os.path.getsize(full_path)
#         file_list.append((file_name, full_path, size))

#     return jsonify(file_list)


import os
import time
from .base import EmbedModelProvider, ChatModelProvider

from latentscope.util import get_key

class OpenAIEmbedProvider(EmbedModelProvider):
    def load_model(self):
        from openai import OpenAI
        import tiktoken
        api_key = get_key("OPENAI_API_KEY")
        if api_key is None:
            print("ERROR: No API key found for OpenAI")
            print("Missing 'OPENAI_API_KEY' variable in:", f"{os.getcwd()}/.env")

        base_url = get_key("OPENAI_BASE_URL")
        if base_url is not None:
            self.client = OpenAI(api_key=api_key, base_url=base_url)
        else:
            self.client = OpenAI(api_key=api_key)

        self.encoder = tiktoken.encoding_for_model(self.name)

    def embed(self, inputs, dimensions=None):
        time.sleep(0.01) # TODO proper rate limiting
        enc = self.encoder
        max_tokens = self.params["max_tokens"]
        inputs = [b.replace("\n", " ") for b in inputs]
        inputs = [enc.decode(enc.encode(b)[:max_tokens]) if len(enc.encode(b)) > max_tokens else b for b in inputs]
        if dimensions is not None and dimensions > 0:
            response = self.client.embeddings.create(
                input=inputs,
                model=self.name,
                dimensions=dimensions
            )
        else:
            response = self.client.embeddings.create(
                input=inputs,
                model=self.name
            )
        embeddings = [embedding.embedding for embedding in response.data]
        return embeddings

class OpenAIChatProvider(ChatModelProvider):
    def load_model(self):
        from openai import OpenAI, AsyncOpenAI
        import tiktoken
        import outlines
        from outlines.models.openai import OpenAIConfig
        if self.base_url is None:
            self.client = AsyncOpenAI(api_key=get_key("OPENAI_API_KEY"))
            self.encoder = tiktoken.encoding_for_model(self.name)
        else:
            self.client = AsyncOpenAI(api_key=get_key("OPENAI_API_KEY"), base_url=self.base_url)
            # even if this is some other model, we wont be able to figure out the tokenizer from custom API
            # so we just use gpt-4o as a fallback, it should be roughly correct for token counts
            self.encoder = tiktoken.encoding_for_model("gpt-4o") 
        config = OpenAIConfig(self.name)
        self.model = outlines.models.openai(self.client, config)
        self.generator = outlines.generate.text(self.model)


    def chat(self, messages):
        response = self.client.chat.completions.create(
            model=self.name,
            messages=messages
        )
        return response.choices[0].message.content

    def summarize(self, items, context=""):
        from .prompts import summarize
        prompt = summarize(items, context)
        return self.generator(prompt)


from .base import ChatModelProvider

class NLTKChatProvider(ChatModelProvider):
    def load_model(self):
        from collections import Counter
        from nltk.corpus import stopwords
        self.Counter = Counter
        self.stopwords = stopwords
        from nltk.tokenize import word_tokenize
        import nltk

        class Encoder():
            def encode(self, text):
                tokens = word_tokenize(text)
                return tokens
            def decode(self, tokens):
                return " ".join(tokens)

        nltk.download('punkt')
        nltk.download('stopwords')
        self.encoder = Encoder()

    # def chat(self, messages):
    #     # TODO: this is kind of hacky, since we aren't really using a chat model
    #     # We are slicing off the first 86 characters because it contains the prompt for LLMs
    #     text = messages[1]["content"][86:]
    #     text = text.replace("<ListItem>", "")
    #     text = text.replace("< ListItem >", "")
    #     text = text.replace("</ListItem>", "")
    #     text = text.replace("</ ListItem >", "")
    #     tokens = self.encoder.encode(text)
    #     # Remove stopwords
    #     tokens = [word for word in tokens if word.isalpha() and word.lower() not in self.stopwords.words('english')]
    #     # Get the top 3 words
    #     top_words = [word for word, count in self.Counter(tokens).most_common(self.params["top_words"])]
    #     label = " ".join(top_words)
    #     return label

    def summarize(self, items, context=""):
        text = "\n".join(items)
        tokens = self.encoder.encode(text)
        tokens = [word for word in tokens if word.isalpha() and word.lower() not in self.stopwords.words('english')]
        # Get the top 3 words
        top_words = [word for word, count in self.Counter(tokens).most_common(self.params["top_words"])]
        label = " ".join(top_words)
        return label



# Usage: python cluster.py <dataset_id> <umap_id> <samples> <min_samples>
# Example: python cluster.py dadabase-curated umap-001 50 5
import os
import re
import sys
import json
import argparse

try:
    # Check if the runtime environment is a Jupyter notebook
    if 'ipykernel' in sys.modules and 'IPython' in sys.modules:
        from tqdm.notebook import tqdm
    else:
        from tqdm import tqdm
except ImportError as e:
    # Fallback to the standard console version if import fails
    from tqdm import tqdm

from latentscope.util import get_data_dir

# TODO move this into shared space
def calculate_point_size(num_points, min_size=10, max_size=30, base_num_points=100):
    import numpy as np
    """
    Calculate the size of points for a scatter plot based on the number of points.
    """
    # TODO fix this to actually calculate a log scale between min and max size
    if num_points <= base_num_points:
        return max_size
    else:
        return min(min_size + min_size * np.log(num_points / base_num_points), max_size)


def main():
    parser = argparse.ArgumentParser(description='Cluster UMAP embeddings')
    parser.add_argument('dataset_id', type=str, help='ID of the dataset')
    parser.add_argument('umap_id', type=str, help='ID of the UMAP file')
    parser.add_argument('samples', type=int, help='Minimum cluster size')
    parser.add_argument('min_samples', type=int, help='Minimum samples for HDBSCAN')
    parser.add_argument('cluster_selection_epsilon', type=float, help='Cluster selection Epsilon', default=0)
    parser.add_argument('column', type=str, nargs='?', help='Use column as cluster labels', default=None)
    
    args = parser.parse_args()
    clusterer(args.dataset_id, args.umap_id, args.samples, args.min_samples, args.cluster_selection_epsilon, args.column)


def clusterer(dataset_id, umap_id, samples, min_samples, cluster_selection_epsilon, column):
    DATA_DIR = get_data_dir()
    cluster_dir = os.path.join(DATA_DIR, dataset_id, "clusters")
    # Check if clusters directory exists, if not, create it
    if not os.path.exists(cluster_dir):
        os.makedirs(cluster_dir)
    # determine the index of the last cluster run by looking in the dataset directory
    # for files named umap-<number>.json
    cluster_files = [f for f in os.listdir(cluster_dir) if re.match(r"cluster-\d+\.json", f)]
    if len(cluster_files) > 0:
        last_cluster = sorted(cluster_files)[-1]
        last_cluster_number = int(last_cluster.split("-")[1].split(".")[0])
        next_cluster_number = last_cluster_number + 1
    else:
        next_cluster_number = 1

    # make the umap name from the number, zero padded to 3 digits
    cluster_id = f"cluster-{next_cluster_number:03d}"
    print("RUNNING:", cluster_id)

    import hdbscan
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    from scipy.spatial import ConvexHull
    from scipy.spatial.distance import cdist

    umap_embeddings_df = pd.read_parquet(os.path.join(DATA_DIR, dataset_id, "umaps", f"{umap_id}.parquet"))
    # Extract x,y columns into numpy array with shape (n,2)
    umap_embeddings = np.column_stack((umap_embeddings_df['x'], umap_embeddings_df['y']))

    if column is not None:
        input_df = pd.read_parquet(os.path.join(DATA_DIR, dataset_id, f"input.parquet"))
        # use the column as the cluster labels
        cluster_labels = input_df[column].to_numpy()
    else:
        clusterer = hdbscan.HDBSCAN(min_cluster_size=samples, min_samples=min_samples, metric='euclidean', cluster_selection_epsilon=cluster_selection_epsilon)
        clusterer.fit(umap_embeddings)
        # Get the cluster labels
        cluster_labels = clusterer.labels_
    # copy cluster labels to another array
    raw_cluster_labels = cluster_labels.copy()

    # Determine points with no assigned cluster
    unique_labels = np.unique(cluster_labels)
    non_noise_labels = unique_labels[unique_labels != -1]
    centroids = [umap_embeddings[cluster_labels == label].mean(axis=0) for label in non_noise_labels]

    # TODO: look into soft clustering
    # https://hdbscan.readthedocs.io/en/latest/soft_clustering.html
    # Assign noise points to the closest cluster centroid
    noise_points = umap_embeddings[cluster_labels == -1]
    if(non_noise_labels.shape[0] > 0):
      closest_centroid_indices = np.argmin(cdist(noise_points, centroids), axis=1)

      # Update cluster_labels with the new assignments for noise points
      noise_indices = np.where(cluster_labels == -1)[0]
      new_assignments = [non_noise_labels[index] for index in closest_centroid_indices]
      cluster_labels[noise_indices] = new_assignments

    
    print("n_clusters:", len(non_noise_labels))
    print("noise points assigned to clusters:", len(noise_points))

    # save umap embeddings to a parquet file with columns x,y
    df = pd.DataFrame({"cluster": cluster_labels, "raw_cluster": raw_cluster_labels})
    output_file = os.path.join(cluster_dir, f"{cluster_id}.parquet")
    df.to_parquet(output_file)
    print(df.head())
    print("wrote", output_file)

    # generate a scatterplot of the umap embeddings and save it to a file
    fig, ax = plt.subplots(figsize=(14.22, 14.22))  # 1024px by 1024px at 72 dpi
    point_size = calculate_point_size(umap_embeddings.shape[0])
    print("POINT SIZE", point_size, "for", umap_embeddings.shape[0], "points")
    plt.scatter(umap_embeddings[:, 0], umap_embeddings[:, 1], s=point_size, alpha=0.5, c=cluster_labels, cmap='Spectral')
    # plot a convex hull around each cluster
    hulls = []
    for label in non_noise_labels:
        indices = np.where(cluster_labels == label)[0]
        points = umap_embeddings[indices]
        # points = umap_embeddings[cluster_labels == label]
        hull = ConvexHull(points)
        hull_list = [indices[s] for s in hull.vertices.tolist()]
        hulls.append(hull_list)
        for simplex in hull.simplices:
            plt.plot(points[simplex, 0], points[simplex, 1], 'k-')

    plt.axis('off')  # remove axis
    plt.gca().set_position([0, 0, 1, 1])  # remove margins
    plt.savefig(os.path.join(cluster_dir, f"{cluster_id}.png"))

    with open(os.path.join(cluster_dir,f"{cluster_id}.json"), 'w') as f:
        json.dump({
            "id": cluster_id,
            "umap_id": umap_id, 
            "samples": samples, 
            "min_samples": min_samples,
            "cluster_selection_epsilon": cluster_selection_epsilon,
            "n_clusters": len(non_noise_labels),
            "n_noise": len(noise_points)
        }, f, indent=2)
    f.close()

    # create the data structure for labeling clusters
    # get the indices of each item in a cluster
    cluster_indices = df.groupby('cluster').groups

    # iterate over the clusters and create a row for each in a new dataframe with a label, description and array of indicies
    slides_df = pd.DataFrame(columns=['label', 'description', 'indices'])
    for cluster, indices in tqdm(cluster_indices.items()):
        label = f"Cluster {cluster}"
        description = f"This is cluster {cluster} with {len(indices)} items."
        new_row = pd.DataFrame({'label': [label], 'description': [description], 'indices': [list(indices)], 'hull': [hulls[cluster]]})
        slides_df = pd.concat([slides_df, new_row], ignore_index=True)

    # write the df to parquet
    slides_df.to_parquet(os.path.join(cluster_dir, f"{cluster_id}-labels-default.parquet"))
    print("done with", cluster_id)

if __name__ == "__main__":
    main()


import json
from .providers.transformers import TransformersEmbedProvider, TransformersChatProvider
from .providers.openai import OpenAIEmbedProvider, OpenAIChatProvider
from .providers.mistralai import MistralAIEmbedProvider, MistralAIChatProvider
from .providers.cohereai import CohereAIEmbedProvider
from .providers.togetherai import TogetherAIEmbedProvider
from .providers.voyageai import VoyageAIEmbedProvider
from .providers.nltk import NLTKChatProvider

# We use a universal id system for models where its:
# <provider>-<model-name> with model-name replacing "/"" with "___"
# i.e. "nomic-ai/nomic-embed-text-v1.5" becomes: 
# "transformers-nomic-ai___nomic-embed-text-v1.5"
# or OpenAI's "text-embedding-3-small" becomes:
# "openai-text-embedding-3-small"

def get_embedding_model_list():
    """Returns a list of available embedding models."""
    from importlib.resources import files
    embedding_path = files('latentscope.models').joinpath('embedding_models.json')
    with open(embedding_path, "r") as f:
        embed_model_list = json.load(f)
    return embed_model_list

def get_embedding_model_dict(id):
    embed_model_list = get_embedding_model_list()
    embed_model_dict = {model['id']: model for model in embed_model_list}
    model = embed_model_dict[id]
    if not model:
        raise ValueError(f"Model {id} not found")
    return model

def get_embedding_model(id):
    """Returns a ModelProvider instance for the given model id."""

    # For backwards compatibility with the old preset transformers models 
    # (all of which were also HF sentence transformers)
    if id.startswith("transformers-"):
        id = id.replace("transformers-", "🤗-")
    if id.startswith("🤗-"):
        # If the model id is a HF sentence transformers model, we get the model id
        # by replacing "/" with "___"
        # Then huggingface will take care of finding the model
        model_name = id.split("🤗-")[1].replace("___", "/")
        model = {
            "provider": "🤗",
            "name": model_name,
            "params": {}
        }
    else:
        model = get_embedding_model_dict(id)
      
    if model['provider'] == "🤗":
        return TransformersEmbedProvider(model['name'], model['params'])
    if model['provider'] == "openai":
        return OpenAIEmbedProvider(model['name'], model['params'])
    if model['provider'] == "mistralai":
        return MistralAIEmbedProvider(model['name'], model['params'])
    if model['provider'] == "cohereai":
        return CohereAIEmbedProvider(model['name'], model['params'])
    if model['provider'] == "togetherai":
        return TogetherAIEmbedProvider(model['name'], model['params'])
    if model['provider'] == "voyageai":
        return VoyageAIEmbedProvider(model['name'], model['params'])


def get_chat_model_list():
    """Returns a list of available chat models."""
    from importlib.resources import files
    chat_path = files('latentscope.models').joinpath('chat_models.json')
    with open(chat_path, "r") as f:
        chat_model_list = json.load(f)
    return chat_model_list

def get_chat_model_dict(id):
    chat_model_list = get_chat_model_list()
    chat_model_dict = {model['id']: model for model in chat_model_list}
    model = chat_model_dict[id]
    if not model:
        raise ValueError(f"Model {id} not found")
    return model

def get_chat_model(id):
    """Returns a ModelProvider instance for the given model id."""
    # For backwards compatibility with the old preset transformers models
    if id.startswith("transformers-"):
        id = id.replace("transformers-", "🤗-")
    if id.startswith("🤗-"):
        # If the model id is a HF model, we get the model name
        # by replacing "/" with "___"
        model_name = id.split("🤗-")[1].replace("___", "/")
        model = {
            "provider": "🤗",
            "name": model_name,
            "params": {}
        }
    elif id.startswith("custom-"):
        # Get custom model from custom_models.json
        import os
        from latentscope.util import get_data_dir
        DATA_DIR = get_data_dir()
        custom_models_path = os.path.join(DATA_DIR, "custom_models.json")
        if os.path.exists(custom_models_path):
            with open(custom_models_path, "r") as f:
                custom_models = json.load(f)
            model = next((m for m in custom_models if m["id"] == id), None)
            if model is None:
                raise ValueError(f"Custom model {id} not found")
        else:
            raise ValueError("No custom models found")
    elif id.startswith("ollama-"):
        model = {
            "provider": "ollama",
            "name": id.split("ollama-")[1],
            "url": "http://localhost:11434/v1", # TODO: this should be passed in somehow?
            "params": {}
        }
    else:
        model = get_chat_model_dict(id)
    
    if model['provider'] == "🤗":
        return TransformersChatProvider(model['name'], model['params'])
    if model['provider'] == "openai":
        return OpenAIChatProvider(model['name'], model['params'])
    if model['provider'] == "custom":
        return OpenAIChatProvider(model['name'], model['params'], base_url=model['url'])
    if model['provider'] == "ollama":
        return OpenAIChatProvider(model['name'], model['params'], base_url=model['url'])
    if model['provider'] == "mistralai":
        return MistralAIChatProvider(model['name'], model['params'])
    if model['provider'] == "nltk":
        return NLTKChatProvider(model['name'], model['params'])




import outlines

summarize_system_prompt = """
You are a helpful assistant who's job it is to summarize lists of items with a short label. 
You will be presented with a list of items and your job is to choose a label that best summarizes the theme of the list so that someone browsing the labels will have a good idea of what is in the list.
"""

@outlines.prompt
def summarize(items, context=""):
  """
You will be presented with a list of items and your job is to summarize the items into a short label of no more than 4 words. 
The items come from a natural language processing task that has grouped them by similaritiy.
Since the items in this list are expected to be similar in some ways, the label will be used to distinguish this list from others, so pay attention to what makes a group of similar items distinct.
{{context}}

Items
--------

{% for item in items%}
<Item>{{ item }}</Item>
{% endfor %}

Task
--------
Choose a label that best summarizes the theme of the list so that someone browsing the labels will have a good idea of what is in the list. 
Do not use punctuation, Do not explain yourself, respond with only a few words that summarize the list.

Label:"""



import re
import os
import sys
import json
import math
import h5py
import logging
import argparse
import numpy as np
import pandas as pd
from importlib.resources import files
from dotenv import dotenv_values, set_key
from flask import Flask, request, jsonify, send_from_directory
from flask_cors import CORS

# from latentscope.util import update_data_dir
from latentscope.util import get_data_dir, get_supported_api_keys
from latentscope.__version__ import __version__


app = Flask(__name__)

app.logger.addHandler(logging.StreamHandler(sys.stderr))
app.logger.setLevel(logging.INFO)

CORS(app, resources={r"/api/*": {"origins": "*"}})

# DATA_DIR = update_data_dir(args.data_dir)
DATA_DIR = get_data_dir()
if not os.path.exists(DATA_DIR):
    os.makedirs(DATA_DIR)

# We enable a read only mode of the server
def check_read_only(s):
    if s is None:
        return False
    return s.lower() in ['true', '1', 't', 'y', 'yes']
# export LATENT_SCOPE_READ_ONLY=1  
READ_ONLY = check_read_only(os.getenv("LATENT_SCOPE_READ_ONLY"))
print("READ ONLY?", READ_ONLY)

# in memory cache of dataframes loaded for each dataset
# used in returning rows for a given index (indexed, get_tags)
DATAFRAMES = {}

from .jobs import jobs_bp, jobs_write_bp
app.register_blueprint(jobs_bp, url_prefix='/api/jobs') 
if(not READ_ONLY):
    app.register_blueprint(jobs_write_bp, url_prefix='/api/jobs') 

from .search import search_bp
app.register_blueprint(search_bp, url_prefix='/api/search') 

from .tags import tags_bp, tags_write_bp
app.register_blueprint(tags_bp, url_prefix='/api/tags') 
if(not READ_ONLY):
    app.register_blueprint(tags_write_bp, url_prefix='/api/tags') 

from .datasets import datasets_bp, datasets_write_bp
app.register_blueprint(datasets_bp, url_prefix='/api/datasets')
if(not READ_ONLY):
    app.register_blueprint(datasets_write_bp, url_prefix='/api/datasets')

from .bulk import bulk_bp, bulk_write_bp
app.register_blueprint(bulk_bp, url_prefix='/api/bulk') 
if(not READ_ONLY):
    app.register_blueprint(bulk_write_bp, url_prefix='/api/bulk') 

from .admin import admin_bp
if not READ_ONLY:
    app.register_blueprint(admin_bp, url_prefix='/api/admin') 

from .models import models_bp, models_write_bp
app.register_blueprint(models_bp, url_prefix='/api/models')
if not READ_ONLY:
    app.register_blueprint(models_write_bp, url_prefix='/api/models')

# ===========================================================
# File based routes for reading data and metadata from disk
# ===========================================================


"""
Allow fetching of dataset files directly from disk
"""
@app.route('/api/files/<path:datasetPath>', methods=['GET'])
def send_file(datasetPath):
    print("req url", request.url)
    return send_from_directory(DATA_DIR, datasetPath)

"""
Given a list of indices (passed as a json array), return the rows from the dataset
"""
# TODO: Should this be deprecated for /query? only used for the hover now
@app.route('/api/indexed', methods=['POST'])
def indexed():
    data = request.get_json()
    dataset = data['dataset']
    indices = data['indices']
    columns = data.get('columns')
    embedding_id = data.get('embedding_id')
    sae_id = data.get('sae_id')

    print("SAE ID", sae_id)

    if dataset not in DATAFRAMES:
        df = pd.read_parquet(os.path.join(DATA_DIR, dataset, "input.parquet"))
        DATAFRAMES[dataset] = df
    else:
        df = DATAFRAMES[dataset]

    if columns:
        df = df[columns]
    
    # get the indexed rows, handling missing indices
    valid_indices = [i for i in indices if i < len(df)]
    rows = df.iloc[valid_indices]
    rows['index'] = valid_indices

    if embedding_id:
        embedding_path = os.path.join(DATA_DIR, dataset, "embeddings", f"{embedding_id}.h5")
        with h5py.File(embedding_path, 'r') as f:
            npvi = np.array(valid_indices)
            sorted_indices = np.argsort(npvi)
            sorted_embeddings = np.array(f["embeddings"][npvi[sorted_indices]])
            filtered_embeddings = sorted_embeddings[np.argsort(sorted_indices)]
        rows['ls_embedding'] = filtered_embeddings
    
    if sae_id:
        sae_path = os.path.join(DATA_DIR, dataset, "saes", f"{sae_id}.h5")
        print("sae_path", sae_path)
        with h5py.File(sae_path, 'r') as f:
            npvi = np.array(valid_indices)
            sorted_indices = np.argsort(npvi)
            sorted_acts = np.array(f["top_acts"][npvi[sorted_indices]])
            filtered_acts = sorted_acts[np.argsort(sorted_indices)]
            sorted_top_inds = np.array(f["top_indices"][npvi[sorted_indices]])
            filtered_top_inds = sorted_top_inds[np.argsort(sorted_indices)]
        # rows['ls_acts'] = filtered_acts
        # rows['ls_top_indices'] = filtered_top_inds
        rows['ls_features'] = [
            {'top_acts': act, 'top_indices': idx} for act, idx in zip(filtered_acts, filtered_top_inds)
        ]

    # send back the rows as json
    return rows.to_json(orient="records")

@app.route('/api/column-filter', methods=['POST'])
def column_filter():
    data = request.get_json()
    dataset = data['dataset']
    filters = data['filters']

    if dataset not in DATAFRAMES:
        df = pd.read_parquet(os.path.join(DATA_DIR, dataset, "input.parquet"))
        DATAFRAMES[dataset] = df
    else:
        df = DATAFRAMES[dataset]
    
    # apply filters
    rows = df.copy()

    print("FILTERS", filters)
    if filters:
        for f in filters:
            if f["type"] == "eq":
                rows = rows[rows[f['column']] == f['value']]
            elif f["type"] == "gt":
                rows = rows[rows[f['column']] > f['value']]
            elif f["type"] == "lt":
                rows = rows[rows[f['column']] < f['value']]
            elif f["type"] == "gte":
                rows = rows[rows[f['column']] >= f['value']]
            elif f["type"] == "lte":
                rows = rows[rows[f['column']] <= f['value']]
            elif f["type"] == "in":
                rows = rows[rows[f['column']].isin(f['value'])]
            elif f["type"] == "contains":
                rows = rows[rows[f['column']].str.contains(f['value'])]

    return jsonify(indices=rows.index.to_list())

@app.route('/api/query', methods=['POST'])
def query():
    per_page = 100
    data = request.get_json()
    dataset = data['dataset']
    page = data['page'] if 'page' in data else 0
    indices = data['indices'] if 'indices' in data else []
    embedding_id = data['embedding_id'] if 'embedding_id' in data else None
    sae_id = data.get('sae_id') if 'sae_id' in data else None

    # filters = data['filters'] if 'filters' in data else None
    sort = data['sort'] if 'sort' in data else None
    if dataset not in DATAFRAMES:
        df = pd.read_parquet(os.path.join(DATA_DIR, dataset, "input.parquet"))
        DATAFRAMES[dataset] = df
    else:
        df = DATAFRAMES[dataset]
    
    # apply filters
    rows = df.copy()
    rows['ls_index'] = rows.index
    

    # get the indexed rows
    if len(indices):
        rows = rows.loc[indices]

    # # only get the first 5 columns
    # cols = 5
    # rows = rows.iloc[:, :cols]

    if embedding_id:
        embedding_path = os.path.join(DATA_DIR, dataset, "embeddings", f"{embedding_id}.h5")
        with h5py.File(embedding_path, 'r') as f:
            npvi = np.array(rows.index)
            sorted_indices = np.argsort(npvi)
            sorted_embeddings = np.array(f["embeddings"][npvi[sorted_indices]])
            filtered_embeddings = sorted_embeddings[np.argsort(sorted_indices)]
        # Add the filtered embeddings as a new column to the rows DataFrame
        rows['ls_embedding'] = filtered_embeddings.tolist()
    
    if sae_id:
        sae_path = os.path.join(DATA_DIR, dataset, "saes", f"{sae_id}.h5")
        print("sae_path", sae_path)
        with h5py.File(sae_path, 'r') as f:
            npvi = np.array(rows.index)
            sorted_indices = np.argsort(npvi)
            sorted_acts = np.array(f["top_acts"][npvi[sorted_indices]])
            filtered_acts = sorted_acts[np.argsort(sorted_indices)]
            sorted_top_inds = np.array(f["top_indices"][npvi[sorted_indices]])
            filtered_top_inds = sorted_top_inds[np.argsort(sorted_indices)]
        # rows['ls_acts'] = filtered_acts
        # rows['ls_top_indices'] = filtered_top_inds
        rows['ls_features'] = [
            {'top_acts': act, 'top_indices': idx} for act, idx in zip(filtered_acts, filtered_top_inds)
        ]

    # print("ROWS", rows.head())
    # apply sort
    if sort:
        rows = rows.sort_values(by=sort['column'], ascending=sort['ascending'])

    # Convert DataFrame to a list of dictionaries
    rows_json = json.loads(rows[page*per_page:page*per_page+per_page].to_json(orient="records"))

    # only send back the first per_page rows
    # per_page = 100
    # rows_json = rows_json[:per_page]

    # send back the rows as json
    return jsonify({
        "rows": rows_json,
        "page": page,
        "per_page": per_page,
        "total": len(rows),
        "totalPages": math.ceil(len(rows) / per_page)
    })

if not READ_ONLY:
    @app.route('/api/settings', methods=['POST'])
    def update_settings():
        data = request.get_json()
        print("update settings", data)
        for key in data:
            set_key(".env", key, data[key])
        return jsonify({})

    @app.route('/api/settings', methods=['GET'])
    def get_settings():
        config = dotenv_values(".env")  # Assuming the .env file is in the root directory
        supported_api_keys = get_supported_api_keys()
        settings = {
            "data_dir": config["LATENT_SCOPE_DATA"],
            "api_keys": [key for key in config if key in supported_api_keys],
            "supported_api_keys": supported_api_keys,
            "env_file": os.path.abspath(".env")
        }
        return jsonify(settings)

@app.route('/api/version', methods=['GET'])
def get_version():
    print("GET VERSION", __version__)
    return __version__


@app.route('/', defaults={'path': ''})
@app.route('/<path:path>')
def catch_all(path):
    print("catch all", path)
    if path.endswith('.js') or path.endswith('.css'):
        pth = files('latentscope').joinpath(f"web/dist/{path}")
        directory = pth.parent
        return send_from_directory(directory, pth.name)
    # always return index.html and let client-side do the routing
    pth = files('latentscope').joinpath("web/dist/index.html")
    directory = pth.parent
    print("sending index.html", directory, pth.name)
    return send_from_directory(directory, pth.name)

def serve(host="0.0.0.0", port=5001, debug=True):
    app.run(host=host, port=port, debug=debug)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Serve the Latent Scope API')
    # parser.add_argument('data_dir', type=str, nargs='?', default=None, help='Path to the directory where data is stored')
    parser.add_argument('--host', type=str, default="0.0.0.0", help='Host to serve the server on')
    parser.add_argument('--port', type=int, default=5001, help='Port to run the server on')
    parser.add_argument('--debug', action='store_true', help='Run server in debug mode')
    args = parser.parse_args()
    host = args.host
    port = args.port
    debug = args.debug
    serve(host, port, debug)


import os
import re
import json
import argparse
from datetime import datetime
from latentscope.util import get_data_dir
from latentscope import __version__


def main():
    parser = argparse.ArgumentParser(description='Setup a scope')
    parser.add_argument('dataset_id', type=str, help='Dataset id (directory name in data folder)')
    parser.add_argument('embedding_id', type=str, help='Embedding id')
    parser.add_argument('umap_id', type=str, help='UMAP id')
    parser.add_argument('cluster_id', type=str, help='Cluster id')
    parser.add_argument('cluster_labels_id', type=str, help='Cluster labels id')
    parser.add_argument('label', type=str, help='Label for the scope')
    parser.add_argument('description', type=str, help='Description of the scope')
    parser.add_argument('--scope_id', type=str, help='Scope id to overwrite existing scope', default=None)
    parser.add_argument('--sae_id', type=str, help='SAE id', default=None)

    args = parser.parse_args()
    scope(**vars(args))

def scope(dataset_id, embedding_id, umap_id, cluster_id, cluster_labels_id, label, description, scope_id=None, sae_id=None):
    DATA_DIR = get_data_dir()
    print("DATA DIR", DATA_DIR)
    directory = os.path.join(DATA_DIR, dataset_id, "scopes")

    def get_next_scopes_number(dataset):
        # figure out the latest scope number
        scopes_files = [f for f in os.listdir(directory) if re.match(r"scopes-\d+\.json", f)]
        if len(scopes_files) > 0:
            last_scopes = sorted(scopes_files)[-1]
            last_scopes_number = int(last_scopes.split("-")[1].split(".")[0])
            next_scopes_number = last_scopes_number + 1
        else:
            next_scopes_number = 1
        return next_scopes_number

    next_scopes_number = get_next_scopes_number(dataset_id)
    # make the umap name from the number, zero padded to 3 digits
    if not scope_id:
        id = f"scopes-{next_scopes_number:03d}"
    else:
        id = scope_id

    print("RUNNING:", id)

    import pandas as pd

    scope = {
        "ls_version": __version__,
        "id": id,
        "embedding_id": embedding_id,
        "umap_id": umap_id,
        "cluster_id": cluster_id,
        "cluster_labels_id": cluster_labels_id,
        "label": label,
        "description": description
    }
    if(sae_id):
        scope["sae_id"] = sae_id

    # read each json file and add its contents to the scope file
    dataset_file = os.path.join(DATA_DIR, dataset_id, "meta.json")
    with open(dataset_file) as f:
        dataset = json.load(f)
        scope["dataset"] = dataset

    embedding_file = os.path.join(DATA_DIR, dataset_id, "embeddings", embedding_id + ".json")
    with open(embedding_file) as f:
        embedding = json.load(f)
        scope["embedding"] = embedding

    if sae_id:
        sae_file = os.path.join(DATA_DIR, dataset_id, "saes", sae_id + ".json")
        with open(sae_file) as f:
            sae = json.load(f)
            scope["sae"] = sae

    umap_file = os.path.join(DATA_DIR, dataset_id, "umaps", umap_id + ".json")
    with open(umap_file) as f:
        umap = json.load(f)
        scope["umap"] = umap
    
    cluster_file = os.path.join(DATA_DIR, dataset_id, "clusters", cluster_id + ".json")
    with open(cluster_file) as f:
        cluster = json.load(f)
        scope["cluster"] = cluster
    
    if cluster_labels_id == "default":
        cluster_labels_id = cluster_id + "-labels-default"
        scope["cluster_labels"] = {"id": cluster_labels_id, "cluster_id": cluster_id}
    else:
        cluster_labels_file = os.path.join(DATA_DIR, dataset_id, "clusters", cluster_labels_id + ".json")
        with open(cluster_labels_file) as f:
            cluster_labels = json.load(f)
            scope["cluster_labels"] = cluster_labels

    # load the actual labels and save everything but the indices in a dict
    cluster_labels_df = pd.read_parquet(os.path.join(DATA_DIR, dataset_id, "clusters", cluster_labels_id + ".parquet"))
    # remove the indices column

    cluster_labels_df = cluster_labels_df.drop(columns=[col for col in ["indices", "labeled", "label_raw"] if col in cluster_labels_df.columns])
    # cluster_labels_df = cluster_labels_df.drop(columns=["indices", "labeled", "label_raw"])
    # change hulls to a list of lists
    cluster_labels_df["hull"] = cluster_labels_df["hull"].apply(lambda x: x.tolist())
    cluster_labels_df["cluster"] = cluster_labels_df.index
    scope["cluster_labels_lookup"] = cluster_labels_df.to_dict(orient="records")
    
    # create a scope parquet by combining the parquets from umap and cluster, as well as getting the labels from cluster_labels
    # then write the parquet to the scopes directory
    umap_df = pd.read_parquet(os.path.join(DATA_DIR, dataset_id, "umaps", umap_id + ".parquet"))
    print("umap columns", umap_df.columns)

    # TODO: make this a shared function with umapper.py
    # or maybe we don't need it in UMAP.py at all?
    def make_tiles(x, y, num_tiles=64):
        import numpy as np
        tile_size = 2.0 / num_tiles  # Size of each tile (-1 to 1 = range of 2)
        
        # Calculate row and column indices (0-63) for each point
        col_indices = np.floor((x + 1) / tile_size).astype(int)
        row_indices = np.floor((y + 1) / tile_size).astype(int)
        
        # Clip indices to valid range in case of numerical edge cases
        col_indices = np.clip(col_indices, 0, num_tiles - 1)
        row_indices = np.clip(row_indices, 0, num_tiles - 1)
        
        # Convert 2D grid indices to 1D tile index (row * num_cols + col)
        tile_indices = row_indices * num_tiles + col_indices
        return tile_indices

    # umap_df['tile_index_32'] = make_tiles(umap_df['x'], umap_df['y'], 32)
    umap_df['tile_index_64'] = make_tiles(umap_df['x'], umap_df['y'], 64)
    # umap_df['tile_index_128'] = make_tiles(umap_df['x'], umap_df['y'], 128)

    cluster_df = pd.read_parquet(os.path.join(DATA_DIR, dataset_id, "clusters", cluster_id + ".parquet"))
    cluster_labels_df = pd.read_parquet(os.path.join(DATA_DIR, dataset_id, "clusters", cluster_labels_id + ".parquet"))
    # create a column where we lookup the label from cluster_labels_df for the index found in the cluster_df
    cluster_df["label"] = cluster_df["cluster"].apply(lambda x: cluster_labels_df.loc[x]["label"])
    print("cluster columns", cluster_df.columns)
    scope_parquet = pd.concat([umap_df, cluster_df], axis=1)
    # TODO: add the max activated feature to the scope_parquet
    # or all the sparse features? top 10?

    print("scope_id", scope_id)
    # create a column to indicate if the row has been deleted in the scope
    scope_parquet["deleted"] = False
    if scope_id is not None:
        # read the transactions file
        transactions_file_path = os.path.join(DATA_DIR, dataset_id, "scopes", scope_id + "-transactions.json")
        with open(transactions_file_path) as f:
            transactions = json.load(f)
            for transaction in transactions:
                if transaction["action"] == "delete_rows":
                    scope_parquet.loc[transaction["payload"]["row_ids"], "deleted"] = True

    # Add an ls_index column that is the index of each row in the dataframe
    scope_parquet['ls_index'] = scope_parquet.index
    print("scope columns", scope_parquet.columns)
    scope_parquet.to_parquet(os.path.join(directory, id + ".parquet"))

    scope["rows"] = len(scope_parquet)
    scope["columns"] = scope_parquet.columns.tolist()
    scope["size"] = os.path.getsize(os.path.join(directory, id + ".parquet"))
    scope["timestamp"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    
    file_path = os.path.join(directory, id + ".json")
    with open(file_path, 'w') as f:
        json.dump(scope, f, indent=2)
    
    transactions_file_path = os.path.join(directory, id + "-transactions.json")
    if not os.path.exists(transactions_file_path):
        with open(transactions_file_path, 'w') as f:
            json.dump([], f)
    
    input_df = pd.read_parquet(os.path.join(DATA_DIR, dataset_id, "input.parquet"))
    input_df.reset_index(inplace=True)
    input_df = input_df[input_df['index'].isin(scope_parquet['ls_index'])]
    combined_df = input_df.join(scope_parquet.set_index('ls_index'), on='index', rsuffix='_ls')
    combined_df.to_parquet(os.path.join(directory, id + "-input.parquet"))

    print("wrote scope", id)

